!+
! Module: epatplot_mod
!
! Authors:  Michael Freyberg - MPE Garching
!           Frank Haberl     - MPE Garching
!           Konrad Dennerl   - MPE Garching
!          
! Description: Plot EPIC pn and EPIC MOS event pattern statistics
!
! Comments:
!
!-

module epatplot_mod

  use cal
  use caloaldefs
  use dal
  use param
  use errorHandling

  implicit none

  private

  public epatplot

contains

!---------------------------------------------------------------------
subroutine epatplot
!---------------------------------------------------------------------

  implicit none

  character(len=10)  ,parameter  :: CodeVersion = '1.1.3'
  integer(kind=Int32),parameter  :: maxbin = 4096, maxbin_times_2 = 2*maxbin
  integer(kind=Int32),parameter  :: maxpat = 4, maxpat_plus_1 = maxpat+1
  integer(kind=Int32),parameter  :: maxmodel = maxpat
! integer             :: statusPG, pgopen
!
  logical(kind=bool)             :: enable_qdp = .false., &
                                    enable_pum = .false., &
                                    enable_y   = .false.
!
  real(kind=single)   ::  h, he, hs, hd, ht, hq, sigma, s,              &
                          xmin, xmax, ymin, ymax, xtxt, ytxt,           &
                          xdat, ydat, xsiz, ysiz,                       &
                          xamin, xamax, xasiz,                          &
                          vxmin, vxmax, vymin, vymax, grey,             &
                          sprb(maxbin), rhis(maxpat,maxbin),            &
                          xhis(1:maxpat,1:maxbin_times_2),              &
		                        yhis(1:maxpat,1:maxbin_times_2),              &
                          model(maxmodel,maxbin), livetime, y, energy(maxbin)
  integer(kind=Int32) ::  ntrow, nsrow, i, j, k, npos = 0,              &
                          nhas = 0, mhas = 0, om1, om2,                 &
                          ix0, ix, iy0, iy1, iy, iex, irow,             &
		                        mbin, kbin, ibin, ks, kd, kt, kq, kpat,       &
                          khis(1:maxpat_plus_1,1:maxbin),               &
		                        jm(maxbin), jp(maxbin),                       &
                          ipat, jpat, jcts(maxbin),                     &
                          icol(maxpat_plus_1), k1, k2, nbin(maxpat)
  integer(kind=Int32) :: CALstat_mode, CALstat_node, CALstat_ccd, &
                         CALstat_inst
  integer(kind=Int32), dimension(2)  :: imageDim
  integer             :: iw, ierr, dthresh, mthresh(0:5),         &
                         calMode, userMode, node = 0, revolut,    &
			                      nccd = 12, iccd, quadrant, ccdid,        &
			                      AccessType = READ, EventsFileType = 0,   &
			                      xw1 = 1, xw2 = 1, nx, ndx,               &
			                      yw1 = 1, yw2 = 1, ny, ndy
  logical(kind=bool)  :: is_ps = .false., is_gif = .false.,       &
                         flag0 = .true., hasCCD(1:12,0:1),        &
			                      is_epn = .false., is_mos = .false., is_rgs = .false.
  logical(kind=bool)  :: xaxisadu = .false., useplotfile = .false.,   &
                         withflag = .true., withoutputmask = .false., &
			                      badpix_from_events = .true.,                 &
			                      offsets_from_events = .true.,                &
			                      hasBADPIX(0:12,0:1) = .false.,               &
			                      hasOFFSET(0:12,0:1) = .false.,               &
			                      withccdnr = .true., singleChip = .false.
! character(len=1)    :: chpt(maxpat)
  character(len=4)    :: ccdmode, mmode(0:5)
  character(len=8)    :: blockname(0:12),blockname2(0:12), revlab
  character(len=40)   :: instmode, device, ext, xlabel, instrument,   &
                         insttxt, rawytxt, filter
  character(len=1024) :: infilename, outfilename, canonicalname,      &
                         outdir, plotfile, outqdpname, outmaskname,   &
			                      omsk1, omsk2
!
  integer(kind=int16),dimension(:,:), pointer     :: FOVmap
  integer(kind=Int8), dimension(:,:), pointer     :: map_arrayP
  integer(kind=Int8), dimension(:,:), allocatable :: map_array

  type(DataSetT)      :: infile, outmask
  type(ArrayT)        :: map
  type(TableT)        :: events, badpix, offset
  type(ColumnT)       :: pattern,pi,flag,rawx,rawy,ccdnr,yextent
  integer(kind=Int8),  dimension(:), pointer :: patternCol,ccdnrCol
  integer(kind=Int16), dimension(:), pointer :: piCol,rawxCol,rawyCol, &
                                                yextentCol
  integer(kind=Int32), dimension(:), pointer :: flagCol
!
  real(kind=single)           ::  conv = 1.0, ypos, userY = 190.0
! real(kind=single),parameter ::  conv = 5.e-3  ! keV/adu
! real(kind=single),parameter ::  conv = 1.0    ! adu channel
  real(kind=single),parameter ::  econv = 5.0   ! eV per adu channel
!
  data icol   / 2, 4, 3, 5, 6 /
  data mthresh  / 40,40,40,40,80,46 /
  data mMode    / 'FF','eFF','LW','SW','TI','BU' /
! data chpt   / 's', 'd', 't', 'q', 'u' /
!
!-----------------------------------------------------------------------
! start the program:
! -------------------------------------------
!
  write(errstr,*) 'epatplot ',trim(CodeVersion),' is running...'
  call message(AppMsg,Sparse)

! -------------------------------------------
! initialize the histograms:
! -------------------------------------------

  khis(1:maxpat_plus_1,1:maxbin) = 0
  rhis(1:maxpat,1:maxbin) = 0.0
  xhis(1:maxpat,1:maxbin_times_2) = 0
  yhis(1:maxpat,1:maxbin_times_2) = 0
  energy(1:maxbin) = 0.0
  model(1:maxmodel,1:maxbin) = 0.0
  jm(1:maxbin) = 0
  jp(1:maxbin) = 0
  jcts(1:maxbin) = 0

! -------------------------------------------
! get input parameters:
! -------------------------------------------

  call getInParamValues(infilename, sigma, device, outdir, withflag, &
                        useplotfile, plotfile, xaxisadu, enable_qdp, &
			                     enable_pum, userMode, enable_y, userY, &
			                     withoutputmask,outmaskname)

  call stripSlash(infilename)
  call getCanonicalName(infilename, canonicalname)
  call stripSlash(canonicalname)
  call stripSlash(plotfile)

  if (withoutputmask) then
    om1 = index(outmaskname,'##')
    om2 = om1+2
    write(errstr,'(1x,a,t35,i9)') 'CCD placeholder ## at position: ',om1
    call message(AppMsg,Noisy)
    if (om1>1) then
      omsk1 = outmaskname(1:om1-1)
      omsk2 = outmaskname(om2:)
    elseif (om1 == 1) then
      omsk1 = ""
      omsk2 = outmaskname(om2:)
    else
      omsk1 = outmaskname
      omsk2 = ""
    endif
  endif

  select case (device)
    case ('/VCPS')
      ext = 'ps'
    case ('/GIF')
      ext = 'gif'
    case default
      ext = 'ps'
  end select

  if (useplotfile) then
    outfilename = trim(outdir) // '/' // trim(plotfile)
    outqdpname  = trim(outdir) // '/' // trim(canonicalname) // '.qdp'
  else
    outfilename = trim(outdir) // '/' // trim(canonicalname) // '.' // trim(ext)
    outqdpname  = trim(outdir) // '/' // trim(canonicalname) // '.qdp'
  endif

  call stripSlash(outfilename)
  call stripSlash(outqdpname)

! print*,'|',trim(outfilename),'|'

  write(errstr,'(1x,a,a)') 'Accessing input data set: ',trim(infilename)
  call message(AppMsg,Verbose)

  if (.not.setexists(infilename)) then
    call error("noSet","Input data set not found")
  endif
  infile = dataSet(infilename,READ,LOW_MEMORY)

  if (.not. hasBlock(infile,'EVENTS')) then    ! no EVENTS extension
    call error("noEvents","No EVENTS extension in input data set")
  endif
  events = table(infile,'EVENTS',READ)

!
! read FILTER attribute and fill relevant text string
!
  if (hasAttribute(infile,"FILTER")) then
    filter = trim(stringAttribute(infile,"FILTER"))
    write(errstr,'(1x,a,t39,a)') 'Filter:     ',trim(filter)
    call message(AppMsg,Verbose)
  else
    filter = ' '
    call message(AppMsg,Verbose,'FILTER attribute not found')
  endif

!
! read INSTRUME attribute and fill relevant text string
!
  if (hasAttribute(events,"INSTRUME")) then
    instrument = trim(stringAttribute(events,"INSTRUME"))
    write(errstr,'(1x,a,t39,a)') 'Instrument: ',trim(instrument)
    call message(AppMsg,Verbose)
  else
    instrument = ' '
    call message(AppMsg,Verbose,'INSTRUME attribute not found')
  endif

  select case (instrument)
    case ('EPN')
      insttxt = 'EPIC pn'
      is_epn = .true.
      CALstat_inst = EPN
    case ('EMOS1')
      insttxt = 'EPIC MOS1'
      is_mos = .true.
      CALstat_inst = EMOS1
    case ('EMOS2')
      insttxt = 'EPIC MOS2'
      is_mos = .true.
      CALstat_inst = EMOS2
    case ('RGS1')
      insttxt = 'RGS1'
      nccd = 9
      is_rgs = .true.
      CALstat_inst = RGS1
    case ('RGS2')
      insttxt = 'RGS2'
      nccd = 9
      is_rgs = .true.
      CALstat_inst = RGS2
    case default
      insttxt = 'EPIC pn'
      is_epn = .true.
      CALstat_inst = EPN
  end select

  if (.not.is_epn) then
    if (xaxisadu) then
      call warning("isnotEPN","xaxisadu=Y only implemented for EPIC pn.")
      xaxisadu = .false.
    endif
  endif
  if (xaxisadu) then
    conv = 1.0
    if (is_epn) then
      xlabel = 'ADU Channel [5 eV]'
    else
      xlabel = 'ADU Channel'
    endif
  else
    conv = 5.0    ! channel -> eV
    xlabel = 'PI Channel [eV]'
  endif

!
! get PI energy of ADU channel
!
  if (is_epn) then
    do i = 1,maxbin
      energy(i) = real(i)*5.00 ! channel -> eV
    enddo
    nccd = 12
    ndx = EPN_MAX_X
    ndy = EPN_MAX_Y
  elseif (is_mos) then
    do i = 1,maxbin
      energy(i) = real(i)*3.25 ! channel -> pseudo-energy
    enddo
    nccd = 7
    ndx = EMOS_MAX_X
    ndy = EMOS_MAX_Y
    mMode(0:5) = 'MOS1'
  else
    call warning("isnotEPIC","only EPIC model curves are implemented yet")
    withoutputmask = .false.   ! only implemented for EPIC at the moment
    do i = 1,maxbin
      energy(i) = real(i)*1.00 ! channel -> pseudo-energy
    enddo
    ndx = 1
    ndy = 1
  endif
  imageDim(1) = ndx
  imageDim(2) = ndy
  nx =  ndx
  ny =  ndy
!
  if (withoutputmask) allocate(map_array(1:ndx,1:ndy))
!
! access events file columns
!
  ntrow = numberOfRows(events)
  if (ntrow .le. 0) then
    call error('emptyInput','No events in input file.')
  else
    write(errstr,'(1x,a,t35,i9)') 'Number of events in input file: ',ntrow
    call message(AppMsg,Verbose)    
  endif

  if (.not.hasColumn(events,'PATTERN')) then
    call error("noPattern","No PATTERN column in input data set")
  endif
  pattern = column(events,'PATTERN',READ)
  patternCol => int8data(pattern)

  if (.not.hasColumn(events,'PI')) then
    call error("noPi","No PI column in input data set")
  endif
  pi = column(events,'PI',READ)
  piCol => int16data(pi)

  if (.not.hasColumn(events,'RAWY')) then
    call error("noPi","No RAWY column in input data set")
  endif
  rawy = column(events,'RAWY',READ)
  rawyCol => int16data(rawy)

  if (.not.hasColumn(events,'FLAG')) then
    if (withflag) then
      call warning("noFlag", &
          "No FLAG column in input data set, using all events for plots.")
      flag0 = .true.
      withflag = .false.
    endif
  else
    flag = column(events,'FLAG',READ)
    flagCol => int32data(flag)
  endif

  if (.not.hasColumn(events,'CCDNR')) then
    if (withccdnr) then
      call warning("noCCDnr", &
          "No CCDNR column in input data set, will check all CCDs.")
      withccdnr = .false.
    endif
  else
    ccdnr = column(events,'CCDNR',READ)
    ccdnrCol => int8data(ccdnr)
  endif

  if (hasAttribute(events,"SUBMODE")) then
    instmode = trim(stringAttribute(events,"SUBMODE"))
  else
    instmode = ' '
    call message(AppMsg,Verbose,'SUBMODE attribute not found')
  endif

  if (hasAttribute(events,"LIVETIME")) then
    livetime = real64Attribute(events,"LIVETIME")
    write(errstr,'(1x,a,t35,f9.1,a)') 'Livetime by evselect: ',livetime,' [s]'
    call message(AppMsg,Verbose)
  else
    livetime = 0.0
  endif

  if (hasAttribute(events,"REVOLUT")) then
    revolut = int16Attribute(events,"REVOLUT")
    write(revlab,'(a,i4.4)') 'Rev.',revolut
  else
    revlab = ' '
  endif

  select case (instmode)
    case ('PrimeFullWindow')
      calMode = 0
      CALstat_Mode = PRIME_FULL_WINDOW
    case ('PrimeFullWindowExtended')
      calMode = 1
      CALstat_Mode = PRIME_FULL_WINDOW_EXTENDED
    case ('PrimeLargeWindow')
      calMode = 2
      CALstat_Mode = PRIME_LARGE_WINDOW
    case ('PrimeSmallWindow')
      calMode = 3
      CALstat_Mode = PRIME_SMALL_WINDOW
    case ('FastTiming')
      calMode = 4
      CALstat_Mode = FAST_TIMING
    case ('FastBurst')
      calMode = 5
      CALstat_Mode = FAST_BURST
    case default
      ccdmode = '  '
      calMode = 0
      CALstat_Mode = PRIME_FULL_WINDOW
      dthresh = 40
  end select

  if (enable_pum) calMode = userMode
  dthresh = mthresh(calMode)
  ccdmode = mmode(calMode)
  write(errstr,'(1x,a,t35,a)') 'Instrument mode of observation:',trim(instmode)
  call message(AppMsg,Verbose)
  write(errstr,'(1x,a,t35,f9.1)') 'Sigma for rebinning: ',sigma
  call message(AppMsg,Verbose)


! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  call message(AppMsg,Noisy,'Reading input data set ...')

  do i = 1, ntrow
    if (withccdnr) then
      iccd = ccdnrCol(i-1)
      if ((iccd.ge.1).and.(iccd.le.nccd)) then
        hasCCD(iccd,node) = .true.
      endif
    endif
    jpat = patternCol(i-1)
    if (withflag) flag0 = (flagCol(i-1) == 0)
    if (jpat.eq.0 .and. flag0) then                       ! singles
      kpat = 1
      ypos = ypos + rawyCol(i-1)
      npos = npos+1
    elseif (jpat.ge.1.and.jpat.le.4 .and. flag0) then     ! doubles
      kpat = 2
      ypos = ypos + rawyCol(i-1)
      npos = npos+1
    elseif (jpat.ge.5.and.jpat.le.8 .and. flag0) then     ! triples
      kpat = 3
      ypos = ypos + rawyCol(i-1)
      npos = npos+1
    elseif (jpat.ge.9.and.jpat.le.12 .and. flag0) then    ! quadruples
      kpat = 4
      ypos = ypos + rawyCol(i-1)
      npos = npos+1
    else
      kpat = 5
    endif
!
! find the energy bin corresponding to event PI channel
!
    kbin = piCol(i-1)/econv                               ! PN only !!

    if (kpat.ge.1 .and. kpat.le.maxpat .and.   &
        kbin.ge.1 .and. kbin.le.maxbin) then
      khis(kpat,kbin) = khis(kpat,kbin) + 1
    endif
  enddo

  write(errstr,'(1x,a,t35,12l1)') 'CCDs in events file: ', &
                (hasCCD(iccd,node),iccd=1,nccd)
  call message(AppMsg,Verbose)

  write(errstr,'(1x,a,t35,i9  )') 'Number of recognized photons: ',npos 
  call message(AppMsg,Verbose)
  if (npos > 0) then
    ypos = ypos/npos
  else
    ypos = 190
  endif
  write(errstr,'(1x,a,t35,f9.3)') 'Average RAWY position: ',ypos 
  call message(AppMsg,Verbose)

  if (enable_Y) then
    write(errstr,'(1x,a,t35,f9.3)') 'User-defined RAWY position: ',userY 
    call message(AppMsg,Verbose)
    ypos = userY
  endif

  if (ypos < 10) then
    write(rawytxt,'(a,f3.1)') 'RAWY = ',ypos
  elseif (ypos < 100) then
    write(rawytxt,'(a,f4.1)') 'RAWY = ',ypos
  else
    write(rawytxt,'(a,f5.1)') 'RAWY = ',ypos
  endif

  if (withflag) call release(flag)
  if (withccdnr) call release(ccdnr)
  call release(pattern)
  call release(pi)
  call release(rawy)

!
! output mask stuff (FLAG==0)
!
  if (withoutputmask) then

    node = 0
    nhas = 0
    mhas = 0
    CALstat_node = node

    hasBADPIX(0:nccd,0:1) = .false.
    hasOFFSET(0:nccd,0:1) = .false.

    do iccd = 0,nccd
      if (iccd == 0) then
        write(blockname(iccd), '(a8)') 'BADPIX  '
        write(blockname2(iccd),'(a8)') 'OFFSETCO'
      else
        write(blockname(iccd), '(a6,i2.2)') 'BADPIX',iccd
        write(blockname2(iccd),'(a6,i2.2)') 'OFFSET',iccd
      endif

      if (hasBlock(infile,blockname(iccd))) then
        hasBADPIX(iccd,node) = .true.
        if (iccd>0) nhas = nhas+1
      endif
      if (hasBlock(infile,blockname2(iccd))) then
        hasOFFSET(iccd,node) = .true.
        if (iccd>0) mhas = mhas+1
      endif
    enddo

    if (hasBADPIX(0,0)) then
      singleChip = .true.
      nhas = 1
    elseif (nhas > 0) then
      singleChip = .false.
    else
      singleChip = .false.
      badpix_from_events = .false.
    endif

    if (singleChip) then
      if (hasBlock(infile,'OFFSETCO')) then      ! old-style single chip file
        EventsFileType = 1
      elseif (hasBlock(infile,'OFFSETS')) then   ! new-style single chip file
        EventsFileType = 2
        blockname2(0) = 'OFFSETS '               ! -> redefine extension name
        hasOFFSET(0,node) = .true.
	       mhas = 1
      else                                       ! very old single chip file
        EventsFileType = 6                       ! -> not supported
        offsets_from_events = .false.
      endif
    else
      if (mhas > 0) then                         ! old-style epchain event file
        EventsFileType = 3
      elseif (hasBlock(infile,'OFFSETS')) then   ! new-style epchain event file
        EventsFileType = 4
        blockname2(1:nccd) = 'OFFSETS '          ! -> redefine extension names
        hasOFFSET(1:nccd,node) = .true.
	       mhas = 1
      else                                       ! old-style LUX/PPS event file
        EventsFileType = 5                       ! -> not supported
        offsets_from_events = .false.
      endif
    endif
    write(errstr,'(1x,a,t35,i9,2x,a,l1,a)') 'Event file type (single): ', &
         EventsFileType,' (',singleChip,')'
    call message(AppMsg,Noisy)
    write(errstr,'(1x,a,t35,i9,i5)') 'Extensions for BADPIX and OFFSET: ', &
         nhas,mhas
    call message(AppMsg,Noisy)

    if (nhas == 0) then    ! no BADPIX extension
      call warning("noBadpix","No BADPIX extension in input data set")
      badpix_from_events = .false.
      call message(AppMsg,Verbose,' Output mask will use BADPIX from CCF instead')
      call message(AppMsg,Verbose,' This is not yet implemented ! No action performed.')
    else
      badpix_from_events = .true.
      call message(AppMsg,Verbose,' Output mask will use BADPIX from events file')
    endif

    if (badpix_from_events) then
      do iccd = 0, nccd
        map_array(1:ndx,1:ndy) = 1
        CALstat_ccd = iccd
        nsrow = 0
        if (hasBADPIX(iccd,node)) then
          badpix = table(infile,blockname(iccd),AccessType)
  	       if (hasAttribute(badpix,"WINDOWX0")) then
  	         xw1 = int16Attribute(badpix,"WINDOWX0")
  	       else
  	         call message(AppMsg,Verbose,'WINDOWX0 attribute not found')
  	       endif
  	       if (hasAttribute(badpix,"WINDOWDX")) then
  	         nx = int16Attribute(badpix,"WINDOWDX")
  	       else
  	         call message(AppMsg,Verbose,'WINDOWDX attribute not found')
  	       endif
  	       xw2 = xw1+nx-1

  	       if (hasAttribute(badpix,"WINDOWY0")) then
  	         yw1 = int16Attribute(badpix,"WINDOWY0")
  	       else
  	         call message(AppMsg,Verbose,'WINDOWY0 attribute not found')
  	       endif
  	       if (hasAttribute(badpix,"WINDOWDY")) then
  	         ny = int16Attribute(badpix,"WINDOWDY")
  	       else
  	         call message(AppMsg,Verbose,'WINDOWDY attribute not found')
  	       endif
  	       yw2 = yw1+ny-1
          write(errstr,'(1x,a,t35,i9,5(2x,i5))') 'Window parameters: ', &
	               xw1,nx,xw2,yw1,ny,yw2
	         call message(AppMsg,Noisy)

	         nsrow = numberOfRows(badpix)
          write(errstr,'(1x,a,t35,i9)') 'Number of bad pixel entries: ', nsrow
	         call message(AppMsg,Noisy)
 	        if (nsrow > 0) then
	           rawx = column(badpix,"RAWX",AccessType)
            rawy = column(badpix,"RAWY",AccessType)
            yextent = column(badpix,"YEXTENT",AccessType)
            rawxCol => int16Data(rawx)
            rawyCol => int16Data(rawy)
            yextentCol => Int16Data(yextent)
          endif
          if (singleChip .and. iccd == 0) then
            if (is_epn) then
  	           quadrant = int8attribute(badpix,"QUADRANT")
	             ccdid = int8attribute(badpix,"CCDID")
	             CALstat_ccd = 3*quadrant+ccdid+1            
              call CAL_setState(CALstat_inst,              &
	                   CALstat_ccd, ccdModeId = CALstat_mode, &   ! CALstat_node, 
		                  accuracyLevel = ACCURACY_HIGH)
            elseif (is_mos) then
	             ccdid = int8attribute(badpix,"CCDID")
	             CALstat_ccd = ccdid            
              call CAL_setState(CALstat_inst,              &
	                   CALstat_ccd, CALstat_node, ccdModeId = CALstat_mode, &
		                  accuracyLevel = ACCURACY_HIGH)
 	          endif
          elseif (.not.singleChip .and. iccd > 0) then
	           call CAL_setState(CALstat_inst,    &
	                CALstat_ccd, ccdModeId = CALstat_mode, &   ! CALstat_node, 
		               accuracyLevel = ACCURACY_HIGH)
          else
            call message(AppMsg,Sparse,' ... BADPIX and BADPIXnn extension ...')
          endif

          call message(AppMsg,Noisy,' ... FLAG mask: BADPIX ...')
  	    	  do irow = 0,nsrow-1
  	    	    iex = yextentCol(irow)
  	    	    ix0 = rawxCol(irow)
  	    	    iy0 = rawyCol(irow)
  	    	    iy1 = min(iy0+iex-1,ndy)
  	    	    map_array(ix0,iy0) = 0
  	    	    do iy = iy0,iy1
  	    	      map_array(ix0,iy) = 0
  	    	      if (ix0.gt.1) map_array(ix0-1,iy) = 0
  	    	      if (ix0.lt.ndx) map_array(ix0+1,iy) = 0
  	    	    enddo
  	    	    if (iy0.gt.1) map_array(ix0,iy0-1) = 0
  	    	    if (iy1.lt.ndy) map_array(ix0,iy1+1) = 0
  	    	  enddo

          if (nsrow > 0) then
    	       call release(rawx)
    	       call release(rawy)
  	         call release(yextent)
	         endif
!
! set FOV map
!
  	    	  call message(AppMsg,Noisy,' ... FLAG mask: FOVmap ...')
  	    	  call CAL_getFOVmap(FOVmap)
  	    	  do iy0 = 1,ndy
  	    	    iy = iy0-1
  	    	    do ix0 = 1,ndx
  	    	      ix = ix0-1
  	    	      map_array(ix0,iy0) = map_array(ix0,iy0)*FOVmap(ix,iy)
  	    	    enddo
  	    	  enddo
  	  !
  	  ! set read-out window map
  	  !
  	    	  call message(AppMsg,Noisy,' ... FLAG mask: WINDOW ...')
  	       if (xw1 .ge. 1 .and. xw2 .ge. xw1 .and. &
  	  	      yw1 .ge. 1 .and. yw2 .ge. yw1 ) then
  	      	    map_array(1:xw1,  1:ndy)   = 0
  	    	    map_array(xw2:ndx,1:ndy)   = 0
  	    	    map_array(1:ndx,  1:yw1)   = 0
  	    	    map_array(1:ndx,  yw2:ndy) = 0
  	  	  else
  	              write(errstr,*) 'Window parameters: XW1,NX,XW2,YW1,NY,YW2 = ', &
  	  	       xw1,nx,xw2,yw1,ny,yw2
  	  	    call warning('invalidWindow')
  	  	  endif
  	  !
  	  ! set offset column map
  	  !
  	      	  call message(AppMsg,Noisy,' ... FLAG mask: OFFSET ...')
  	  if (hasOFFSET(iccd,node)) then
  	    offset = table(infile,blockname2(iccd),AccessType)
  	    nsrow = numberOfRows(offset)
            write(errstr,'(1x,a,t35,i9)') 'Number of offset columns: ', nsrow
	    call message(AppMsg,Noisy)
	    if (nsrow > 0) then
  	      rawx = column(offset,"RAWX",AccessType)
  	      rawxCol => int16Data(rawx)
              ccdnr = column(offset,'CCDNR',READ)
              ccdnrCol => int8data(ccdnr)

              do irow = 0,nsrow-1
	        if (ccdnrCol(irow) == CALstat_ccd) then
    	          ix = rawxCol(irow)
                  map_array(ix,1:ndy) = 0
		  if (ix.gt.1)   map_array(ix-1,1:ndy) = 0
		  if (ix.lt.ndx) map_array(ix+1,1:ndy) = 0
		endif
              enddo

      	      call release(rawx)     ! column
    	      call release(ccdnr)    ! column
	    endif
    	    call release(offset)   ! extension
          else
	    write(errstr,*) 'Does not have OFFSET for CCD = ',iccd,' [',&
	                     trim(blockname2(iccd)),']'
            call message(AppMsg,Noisy)   
    	  endif
 
    	  write(outmaskname,'(a,i2.2,a)') trim(omsk1),CALstat_ccd,trim(omsk2)
          write(errstr,*) '... FLAG mask: ',trim(outmaskname)
    	  call message(AppMsg,Verbose)
    	  outmask = dataSet(outmaskname,Create)
    	  map = addArray(outmask, "PRIMARY", Integer8, imageDim)
   	  map_arrayP => int8Array2Data(map)
   	  map_arrayP = map_array
   	  call setAttributes(attributable(map),attributable(badpix))
          if (hasAttribute(map,"HISTORY")) then
	    call deleteAttribute(map,"HISTORY")
	  else
	    call message(AppMsg,Noisy,' ... No HISTORY keyword.')
	  endif
          if (hasAttribute(map,"COMMENT")) then
	    call deleteAttribute(map,"COMMENT")
	  else
	    call message(AppMsg,Noisy,' ... No COMMENT keyword.')
	  endif
   	  call release(map)
   	  call release(outmask)
  	  call release(badpix)
        endif ! hasBADPIX
      enddo   ! loop over iccd=0,nccd
    endif     ! badpix_from_events
    if (.not. offsets_from_events) then
      call warning('noOFFSET','Cannot reliably handle OFFSETS, not included in FLAG map.')
    endif
  endif       ! outputmask requested

!
  call release(events)
  call release(infile)

  if (allocated(map_array)) deallocate(map_array)
!
!-----------------------------------------------------------------------
!
!     accumulate the distributions
!
  call message(AppMsg,Sparse,' ')
  call message(AppMsg,Verbose,' Pattern statistics for recognized events:')

  ks = 0
  kd = 0
  kt = 0
  kq = 0
  ymax = 0.0
  ymin = 1.0e+7
!
  do ibin = 1, maxbin
    ks = ks + khis(1,ibin)
    kd = kd + khis(2,ibin)
    kt = kt + khis(3,ibin)
    kq = kq + khis(4,ibin)
    khis(maxpat_plus_1,ibin) = &
                khis(1,ibin) + khis(2,ibin) + khis(3,ibin) + khis(4,ibin)
  enddo
!
  k = ks + kd + kt + kq
!
  hs = real(ks) / real(k)
  hd = real(kd) / real(k)
  ht = real(kt) / real(k)
  hq = real(kq) / real(k)
!
  write(errstr,'(4(7x,a1))') 's','d','t','q'
  call message(AppMsg,Sparse)
  write(errstr,'(4f8.4)') hs,hd,ht,hq
  call message(AppMsg,Sparse)

!
! compute the pattern distributions vs. channel:
!
  call message(AppMsg,Noisy,' ... compute pattern distributions ...')
  do ipat = 1, maxpat
    h = sigma * sigma
    s = 0.0
    i = 1
    j = 0
    k = 0
    do while (khis(ipat,i).eq.0 .and. i.lt.maxbin)
      i = i + 1
    enddo
    jm(1) = i
    if (i.gt.1) i = i - 1
    do while (i.lt.maxbin)
      i = i + 1
      s = s + real(khis(ipat,i))
      if (s*s.gt.h*s) then
        k = k + 1
        jcts(k) = s
        jp(k) = i
        s = 0.0
      endif
    enddo
    mbin = k

    if (mbin > 0) then
      do i = 2, mbin
        jm(i) = jp(i-1) + 1
      enddo

      do i = 1, mbin
        h = real(jp(i)-jm(i)) + 1.0
        sprb(i)  = real(jcts(i)) / h
      enddo
!
! store as histogram
!
      j = j+1
      xhis(ipat,j) = alog10((real(jm(1))-0.5)*conv)
      y = alog10(sprb(1))
      yhis(ipat,j) = y
      j = j+1
      xhis(ipat,j) = alog10((real(jp(1))+0.5)*conv)
      yhis(ipat,j) = y
      ymax = max(ymax,y)
      ymin = min(ymin,y)
      do i = 2, mbin
        j = j+1
        xhis(ipat,j) = xhis(ipat,j-1) 
        y = alog10(sprb(i))
        yhis(ipat,j) = y
        j = j+1
        xhis(ipat,j) = alog10((real(jp(i))+0.5)*conv)
        yhis(ipat,j) = y
        ymax = max(ymax,y)
        ymin = min(ymin,y)
      enddo
    else
      write(errstr,'(a,i1)') &
          'not enough statistics for at least 1 bin for PAT = ',ipat
      call warning('sigmaTooLarge')
    endif
    nbin(ipat) = j
  enddo

!
! initialize the plot for pattern distributions:
!
  call message(AppMsg,Noisy,' ')
  call message(AppMsg,Noisy,' ... prepare plot input data set ...')

  if (index(trim(device),"PS") .gt. 0)  is_ps = .true.
  if (index(trim(device),"GIF") .gt. 0) is_gif = .true.

! statusPG = pgopen(trim(device))
  if (is_ps) then
    call pgbegin(0,trim(outfilename)//trim(device),1,1)
  elseif (is_gif) then
    call pgbegin(0,trim(outfilename)//trim(device),1,1)
  else
    call pgbegin(0,trim(device),1,1)
  endif
  call pgpaper(7.8,1.414)
!
! top bar with some text
!
  vxmin = 0.15 - 0.75*0.12
  vxmax = 0.90
  vymin = 0.98
  vymax = 1.00
  grey  = 0.95
  call pgscr(255,grey,grey,grey)
  call pgvport(vxmin,vxmax,vymin,vymax)
  call pgwindow(0.0,1.0,0.0,1.0)
  call pgsci(255)
  call pgrect(0.0,1.0,0.0,1.0)
  call pgsci(1)
  call pgsch(0.5)
  call pgslw(1)
  call pgptext(0.99, 0.4, 0.0, 1.0, trim(outfilename))
  call pgptext(0.01, 0.4, 0.0, 0.0, & 
           'XMM '//trim(insttxt)// &
           ', pattern statistics: epatplot-'//trim(CodeVersion))
!
! username, date and time at bottom of plot
!
  call pgsch(1.0)
  call pgiden

  call message(AppMsg,Noisy,' ... prepare X and Y ranges ...')
  xmin = alog10(17.0*conv)
  xmax = alog10(4096.0*conv)
  xamin = alog10(17.0)
  xamax = alog10(4096.0)
!  ymin = max(alog10(1.e-3),real(int(alog10(ymin)-1)))
!  ymax = real(int(alog10(ymax)+1))
  ymin = int(max(-3.0,ymin))-1
  ymax = int(ymax)+1
  write(errstr,*) '... unreasonable ranges: XMIN,XMAX, YMIN,YMAX = ', &
     xmin,xmax,ymin,ymax
  if (ymin>ymax) then
    call message(AppMsg,Verbose)
    call message(AppMsg,Verbose,' ... YMIN has been set to YMAX-2 ...')
    ymin = ymax - 2
  else
    call message(AppMsg,Noisy+1)  
  endif

  xsiz = xmax - xmin
  xasiz = xamax - xamin
  ysiz = ymax - ymin
  vxmin = 0.15
  vxmax = 0.90
  vymin = 0.52
  vymax = 0.90
  call message(AppMsg,Noisy+1,' ... call pgvport(vxmin,vxmax,vymin,vymax) ...')
  call pgvport(vxmin,vxmax,vymin,vymax)
  call message(AppMsg,Noisy+1,' ... call pgwindow(xmin,xmax,ymin,ymax) ...')
  call pgwindow(xmin,xmax,ymin,ymax)
  call message(AppMsg,Noisy+1,' ... call pgsci(255) ...')
  call pgsci(255)
  call message(AppMsg,Noisy+1,' ... call pgrect(xmin,xmax,ymin,ymax) ...')
  call pgrect(xmin,xmax,ymin,ymax)
  call message(AppMsg,Noisy+1,' ... call pgsci(1) ...')
  call pgsci(1)
  call pgslw(3)
  call pgbox('BC',0.0,0,'BC',0.0,0)
  call pgbox('LN',0.0,0,' ',0.0,0)
  call pgsch(0.75)
  call pgbox(' ',0.0,0,'LNV',0.0,0)
  call pgsch(1.0)
  call pgslw(1)
  call pgbox('BCLST',0.0,0,'BCLST',0.0,0)
  call pgslw(5)
  call pgsch(1.1)
  xtxt = (xmax + xmin)*0.5
  ytxt = ymin - ysiz*0.11
  call message(AppMsg,Noisy+1,' ... call pgptext(xtxt, ytxt, 0.0, 0.5, trim(xlabel)) ...')
  call pgptext(xtxt, ytxt, 0.0, 0.5, trim(xlabel))
! call pgsch(1.25)
  call pgsch(1.12)
  call pgslw(4)
  xtxt = xmin - xsiz*0.12
  ytxt = ymax + ysiz*0.12
  call message(AppMsg,Noisy+1,' ... call pgptext(INSTTXT) ...')
  call pgptext(xtxt, ytxt, 0.0, 0.0, 'XMM / '//trim(insttxt)//' '//trim(instmode)//' '//trim(filter))
  xtxt = xmax
  ytxt = ymax + ysiz*0.12
  call pgsch(0.65)
  call pgslw(3)
  call pgptext(xtxt, ytxt, 0.0, 1.0, trim(revlab))
! call pgsch(1.0)
  call pgslw(3)
  call pgsch(0.65)
  ytxt = ymax + ysiz*0.05
  xtxt = xmax
  call message(AppMsg,Noisy+1,' ... call pgptext(FILENAME) ...')
  call pgptext(xtxt, ytxt, 0.0, 1.0, trim(infilename))
  call pgsch(1.0)
  call pgslw(4)
  call pgsci(5)
  xtxt = xmin - xsiz*0.1
  ytxt = ymin + ysiz*0.125
  call pgptext(xtxt, ytxt, 90.0, 0.5, 'q/adu')
  call pgsci(3)
  ytxt = ytxt + ysiz*0.25
  call pgptext(xtxt, ytxt, 90.0, 0.5, 't/adu')
  call pgsci(4)
  ytxt = ytxt + ysiz*0.25
  call pgptext(xtxt, ytxt, 90.0, 0.5, 'd/adu')
  call pgsci(2)
  ytxt = ytxt + ysiz*0.25
  call pgptext(xtxt, ytxt, 90.0, 0.5, 's/adu')
  call pgsci(1)
  call pgslw(1)
!
! plot the pattern distributions vs. channel:
!
  call message(AppMsg,Noisy,' ... plot pattern distributions ...')
  do ipat = 1, maxpat
    j = 0
!
! plot as histogram
!
    if (nbin(ipat) .gt. 0) then
      call pgsci(icol(ipat))
      j = j+1
      xdat = xhis(ipat,j)
      ydat = yhis(ipat,j)
      call pgmove(xdat,ydat)
      j = j+1
      xdat = xhis(ipat,j)
      call pgdraw(xdat,ydat)
      do i = j+1, nbin(ipat), 2
        ydat = yhis(ipat,i)
        call pgdraw(xdat,ydat)
        xdat = xhis(ipat,i+1)
        call pgdraw(xdat,ydat)
      enddo
    endif
  enddo
!
! initialize the plot for pattern statistics:
!
  call message(AppMsg,Noisy,' ... plot pattern statistics ...')
  ymin = -0.05
  ymax = 1.05
  ysiz = ymax - ymin
  vymin = vymin - 0.45
  vymax = vymax - 0.45
  call pgvport(vxmin,vxmax,vymin,vymax)
  call pgwindow(xmin,xmax,ymin,ymax)
  call pgsci(255)
  call pgrect(xmin,xmax,ymin,ymax)
  call pgsci(1)
  call pgslw(3)
  call pgbox('BC',0.0,0,'BC',0.0,0)
  call pgbox('LN',0.0,0,' ',0.0,0)
  call pgsch(0.75)
  call pgbox(' ',0.0,0,'NV',0.0,0)
  call pgsch(1.0)
  call pgslw(1)
  call pgbox('A',0.0,0,' ',0.0,0)
  call pgbox('BCLST',0.0,0,'BCST',0.0,0)
! call pgslw(6)
  call pgslw(5)
  call pgsch(1.1)
  xtxt = (xmax + xmin)*0.5
  ytxt = ymin - ysiz*0.11
  call pgptext(xtxt, ytxt, 0.0, 0.5, trim(xlabel))
  call pgslw(3)
  call pgsch(0.7)
  xtxt = xmin + xsiz*0.94
  ytxt = ymin + ysiz*0.85
  call pgptext(xtxt, ytxt, 0.0, 1.0, trim(RAWYTXT))
  call pgslw(5)
  call pgsch(1.1)
  xtxt = xmin + xsiz*0.95
  ytxt = ymin + ysiz*0.91
  if (ccdmode(1:2) .ne. "  ") then
!   if (calMode .ne. 5) then
      call pgptext(xtxt, ytxt, 0.0, 1.0, trim(ccdmode)//' Model')
!   else
!     call pgptext(xtxt, ytxt, 0.0, 1.0, trim(ccdmode)//' (no model yet)')
!   endif
  endif
  call pgslw(4)
  call pgsci(icol(4))
  xtxt = xmin - xsiz*0.1
  ytxt = ymin + ysiz*0.125
  call pgptext(xtxt, ytxt, 90.0, 0.5, 'q')
  call pgsci(icol(3))
  ytxt = ytxt + ysiz*0.25
  call pgptext(xtxt, ytxt, 90.0, 0.5, 't')
  call pgsci(icol(2))
  ytxt = ytxt + ysiz*0.25
  call pgptext(xtxt, ytxt, 90.0, 0.5, 'd')
  call pgsci(icol(1))
  ytxt = ytxt + ysiz*0.25
  call pgptext(xtxt, ytxt, 90.0, 0.5, 's')
  call pgsci(1)
  call pgslw(1)
!
! determine binning for statistics plot
!
  call message(AppMsg,Noisy,' ... determine binning for statistics plot ...')
  ipat = 5
  h = sigma * sigma * 10.0  * 10.0
  s = 0.0
  i = 1
  k = 0
  do while (khis(ipat,i).eq.0 .and. i.lt.maxbin)
    i = i + 1
  enddo
  jm(1) = i
  if (i.gt.1) i = i - 1
  do while (i.lt.maxbin)
    i = i + 1
    s = s + real(khis(ipat,i))
    if (s*s.gt.h*s) then
      k = k + 1
      jp(k) = i
      s = 0.0
    endif
  enddo
  mbin = k
  do i = 2, mbin
    jm(i) = jp(i-1) + 1
  enddo
!
! determine the fraction of patterns among the recognized photons:
!
  call message(AppMsg,Noisy,&
    ' ... determine the fraction of patterns among the recognized photons ...')
  do ipat = 1, maxpat
    do i = 1, mbin
      k1 = 0
      k2 = 0
      do ibin = jm(i), jp(i)
        k1 = k1 + khis(ipat,ibin)
        k2 = k2 + khis(5,ibin)
      enddo
      rhis(ipat,i) = real(k1) / amax1(real(k2),1.0)
    enddo
!
! plot as histogram
!
    call pgsci(icol(ipat))
    xdat = alog10((real(jm(1))-0.5)*conv)
    ydat = rhis(ipat,1)
    call pgmove(xdat,ydat)
    xdat = alog10((real(jp(1))+0.5)*conv)
    call pgdraw(xdat,ydat)
    do i = 2, mbin
      ydat = rhis(ipat,i)
      call pgdraw(xdat,ydat)
      xdat = alog10((real(jp(i))+0.5)*conv)
      call pgdraw(xdat,ydat)
    enddo
  enddo
!------------------------------------------------------------------------------
! plot the standard single and double pattern statistics 
! as used in the response matrix
!
  call message(AppMsg,Noisy,' ... model curve as used for DRM ...')
  do ipat = 1,2
    model(ipat,1:maxbin) = 0.0
    if (is_epn) then
      call patmod(model(ipat,1:maxbin), maxbin, calMode, dthresh, ipat, ypos)
    else
      call patmod_mos(model(ipat,1:maxbin), maxbin, energy, ipat)
    endif
  enddo

!
! plot as line
!
  call message(AppMsg,Noisy,' ... plot as line ...')
  if (is_epn) then
    call pgwindow(xamin,xamax,ymin,ymax)
    do ipat = 1,2
      call pgsci(icol(ipat))
      i = 20
      xdat = alog10(real(i))
      ydat = model(ipat,i)
      call pgmove(xdat,ydat)
      do i = 21, maxbin
      	 xdat = alog10(real(i))
  	     ydat = model(ipat,i)
  	     call pgdraw(xdat,ydat)
      enddo
    enddo
  else
    do ipat = 1,2
      call pgsci(icol(ipat))
      i = 2
!     call message(AppMsg,Noisy,' ... !EPN: starting point ...')
      xdat = alog10(energy(i))
      ydat = model(ipat,i)
      call pgmove(xdat,ydat)
      do i = 3, maxbin
        if (energy(i) > 0) then
  	       xdat = alog10(energy(i))
  	       ydat = model(ipat,i)
!	        print*,i,energy(i),xdat,ydat
  	       call pgdraw(xdat,ydat)
	       endif
      enddo
!     call message(AppMsg,Noisy,' ... !EPN: loop done...')
    enddo
  endif
  call message(AppMsg,Noisy,' ... call pgend ...')
  call pgend
!
! write out results into ASCII file which can be plotted with QDP
!
  if (enable_qdp) then
    call get_io_Unit(iw)
    open(iw,file=outqdpname,status='unknown',iostat=ierr)
    if (ierr .eq. 0) then
      write(iw,'(a,a)') '! ',trim(outqdpname)
      write(iw,'(a)')'read serr 1'
      write(iw,'(a)')'col 0 on 2 3'
      write(iw,'(a)')'label x ADU Channel'
      write(iw,'(a)')'label y Pattern fraction'
      write(iw,'(a)')'resc y -0.05 1.05'
      write(iw,'(a)')'log x'
      write(iw,'(a)') '! CH           CHmin  CHmax  s'//&
                                  '        d        t        q'
      write(iw,'(a)') '! -----        -----  -----  ------   '//&
                                    '------   ------   ------'
  !
      do i = 1, mbin
        h = (jp(i)+jm(i))*0.5
        he = (jp(i)-jm(i))*0.5
        write(iw,'(2f6.1,2i7,4f9.4)') h,he,jm(i),jp(i),&
            rhis(1,i),rhis(2,i),rhis(3,i),rhis(4,i)
      enddo
      close(iw)
      call message(AppMsg,Verbose,' ')
      write(errstr,*) 'QDP output written to file ' // trim(outqdpname)
    else
      write(errstr,*) 'QDP output file could not be opened:' // trim(outqdpname)
    endif
    call message(AppMsg,Verbose)
  else
    call message(AppMsg,Verbose,' ')
  endif

  write(errstr,*) 'PostScript output written to file ' // trim(outfilename)
  call message(AppMsg,Sparse)
  call message(AppMsg,Verbose,' ')

end subroutine epatplot


! =========================
! Program parameter reading
! =========================
subroutine getInParamValues(infilename, sigma, device, outdir, withflag, &
                            useplotfile, plotfile, xaxisadu, enable_qdp, &
			                         enable_pum, userMode, enable_Y, userY, &
			                         withoutputmask,outmaskname)

  integer            ,intent(out)   :: userMode
  real(kind=single)  ,intent(out)   :: sigma, userY
  logical(kind=bool) ,intent(out)   :: useplotfile,xaxisadu,enable_qdp,&
                                       withflag, enable_pum, enable_Y, &
				                                   withoutputmask
  character(len=*)   ,intent(out)   :: infilename,device,outdir,plotfile,&
                                       outmaskname

  infilename  = trim(stringParameter('set'))
  sigma       = realParameter('sigma')
  userMode    = intParameter('usermode')
  enable_pum  = booleanParameter('withusermode')
  userY       = realParameter('userrawy')
  enable_Y    = booleanParameter('withuserrawy')
  withflag    = booleanParameter('withflag')
  withoutputmask = booleanParameter('withoutputmask')
  outmaskname = stringParameter("outmaskname")
  useplotfile = booleanParameter('useplotfile')
  device      = trim(stringParameter('device'))
  outdir      = trim(stringParameter('outdir'))
  plotfile    = trim(stringParameter('plotfile'))
  xaxisadu    = booleanParameter('xaxisadu')
  enable_qdp  = booleanParameter('withqdp')

  if (verbosity(AppMsg) > Noisy) then
    call message(AppMsg,Sparse,'List of parameters as read:')
    write(errstr,*) 'set             -> infilename        ',infilename
    call message(AppMsg,Sparse)
    write(errstr,*) 'sigma           -> sigma	  	  ',sigma
    call message(AppMsg,Sparse)
    write(errstr,*) 'usermode        -> userMode  	  ',userMode
    call message(AppMsg,Sparse)
    write(errstr,*) 'withusermode    -> enable_pum	  ',enable_pum
    call message(AppMsg,Sparse)
    write(errstr,*) 'userrawy        -> userY	  	  ',userY
    call message(AppMsg,Sparse)
    write(errstr,*) 'withuserrawy    -> enable_Y  	  ',enable_Y
    call message(AppMsg,Sparse)
    write(errstr,*) 'withflag        -> withflag  	  ',withflag
    call message(AppMsg,Sparse)
    write(errstr,*) 'withoutputmask  -> withoutputmask	  ',withoutputmask
    call message(AppMsg,Sparse)
    write(errstr,*) 'outmaskname     -> outmaskname	  ',outmaskname
    call message(AppMsg,Sparse)
    write(errstr,*) 'useplotfile     -> useplotfile	  ',useplotfile
    call message(AppMsg,Sparse)
    write(errstr,*) 'device          -> device	 	  ',device
    call message(AppMsg,Sparse)
    write(errstr,*) 'outdir          -> outdir	 	  ',outdir
    call message(AppMsg,Sparse)
    write(errstr,*) 'plotfile        -> plotfile 	  ',plotfile
    call message(AppMsg,Sparse)
    write(errstr,*) 'xaxisadu        -> xaxisadu   	  ',xaxisadu
    call message(AppMsg,Sparse)
    write(errstr,*) 'withqdp         -> enable_qdp        ',enable_qdp
    call message(AppMsg,Sparse)
  endif

! if (.NOT. xaxisadu) then
!   call message(AppMsg,Sparse,'Parameter setting XAXISADU=N not ' // &
!               'implemented yet, using XAXISADU=Y in the following.')
!   xaxisadu = .true.
! endif

end subroutine getInParamValues


! ==========================================
! Create output filename from input filename
! ==========================================
subroutine getCanonicalName(infilename,canonicalname)

  implicit none
  character(len=*)   ,intent(in)    :: infilename
  character(len=*)   ,intent(out)   :: canonicalname
  integer                           :: ns,nd,nl,il

  nl = len(trim(infilename))
  do il = nl, 1, -1
    if (infilename(il:il) == '/') exit
  enddo
  ns = max(il,1)
! print*,'IL,NS = ',il,ns
  do il = nl, ns, -1
    if (infilename(il:il) == '.') exit
  enddo
  if (il <= ns) then
    nd = nl
  else
    nd = max(il,2)-1
  endif
! print*,'IL,ND = ',il,nD
  canonicalname = infilename(ns:nd) // '_pat'
! print*,'|',trim(canonicalname),'|'

end subroutine getCanonicalName


! ==========================================
! Create output filename from input filename
! ==========================================
subroutine stripSlash(infilename)

  implicit none
  character(len=*)   ,intent(inout) :: infilename
  integer                           :: nl,il,jl

  nl = len(trim(infilename))
  do il = nl, 1, -1
    if (infilename(il:il) == '/') then
      do jl = il-1, 1, -1
        if (infilename(jl:jl) == '/') then
	         infilename(jl:nl) = infilename(jl+1:nl)//' '
	       else
	         exit
	       endif
      enddo
    endif
  enddo
  nl = len(trim(infilename))
  do il = nl, 3, -1
    if (infilename(il-2:il) == '/./') then
      infilename(il-2:nl) = infilename(il:nl)//'  '
    endif
  enddo
  if (infilename(1:2) == './') then
    infilename(1:nl) = infilename(3:nl)//'  '
  endif
! print*,'|',trim(infilename),'|'

end subroutine stripSlash


! ==========================================================================
! suggested by Clive Page in SSC-LUX-TN-0005 (Issue 1.0, 1996 Sept 2)
! ==========================================================================
subroutine get_io_unit(iounit)

  implicit none
  integer, intent(out) :: iounit
  logical              :: lopen, lexist
  do iounit = 99,0,-1
    inquire(unit=iounit, exist=lexist, opened=lopen)
    if (lexist .and. .not. lopen) exit
  enddo
  
end subroutine get_io_unit


! ==========================================================================
! returns the pattern distribution model
! ==========================================================================
subroutine patmod(patfraction, maxbin, calMode, dthresh, pattyp, line)

  implicit none
  integer          ,intent(in)    :: calMode, dthresh, pattyp, maxbin
  real(kind=single),intent(in)    :: line
  real(kind=single),intent(out)   :: patfraction(1:maxbin)
  integer                         :: ibin
  real(kind=single)               :: rbin

  integer  	                      :: ndthresh, irun = 0
  integer  	                         ipar, iy1, iy2
  real  	                            rly1, rly2
! character*2	ccdmode
  integer      	maxff, maxef, maxsw, maxlw, maxti, maxbu, npar, ntyp
  parameter   ( maxff=10, maxef=10, maxsw=3, maxlw=5, maxti=1, maxbu=1 )
  parameter   ( npar=9, ntyp=2 )
  real  	ffpar(npar,maxff,ntyp), efpar(npar,maxef,ntyp)
  real  	swpar(npar,maxsw,ntyp), lwpar(npar,maxlw,ntyp)
  real  	tipar(npar,maxti,ntyp), bupar(npar,maxbu,ntyp)
  real  	par(npar)

! FullFrame: parameters for focus region (Y9) derived from  MS1229.2+6430, rev 0082
!                       for the other regions taken from ext. FF
  data ffpar/                                                           &
       0.5284, 1385.9, 0.1524, 158.155, 0.1762, 19.5666, 0.1435, 31.05, 10., &
       0.5284, 1385.9, 0.1524, 158.155, 0.1762, 19.5666, 0.1435, 31.93, 30., &
       0.5284, 1385.9, 0.1524, 158.155, 0.1762, 19.5666, 0.1435, 32.84, 50., &
       0.5354, 1394.1, 0.1342, 200.864, 0.1629, 30.6517, 0.1627, 33.77, 70., &
       0.5320, 1499.4, 0.1311, 222.169, 0.1601, 35.5882, 0.1747, 34.73, 90., &
       0.5334, 1725.7, 0.1222, 260.906, 0.1489, 42.7871, 0.1926, 35.71,110., &
       0.5553, 1184.2, 0.1230, 190.263, 0.1669, 31.6242, 0.1550, 36.73,130., &
       0.5650, 955.74, 0.1335, 163.263, 0.1682, 25.0995, 0.1349, 37.77,150., &
       0.5649, 943.63, 0.1404, 161.558, 0.1691, 24.2733, 0.1295, 38.84,170., &
       0.5331, 1329.,  0.1544, 180.6,	  0.2007, 34.89,   0.1181, 40.00,190., &
       0.4101, 1356.3,-0.1121, 142.936,-0.1630, 19.8420,-0.1355, 31.05, 10., &
       0.4101, 1356.3,-0.1121, 142.936,-0.1630, 19.8420,-0.1355, 31.93, 30., &
       0.4101, 1356.3,-0.1121, 142.936,-0.1630, 19.8420,-0.1355, 32.84, 50., &
       0.4047, 1462.5,-0.0942, 193.856,-0.1428, 32.6295,-0.1629, 33.77, 70., &
       0.4015, 1484.8,-0.0849, 224.818,-0.1348, 38.8510,-0.1753, 34.73, 90., &
       0.4283, 2242.7,-0.1044, 259.620,-0.1286, 43.3512,-0.1936, 35.71,110., &
       0.3867, 987.81,-0.0955, 142.543,-0.1753, 15.0092,-0.1149, 36.73,130., &
       0.3812, 870.98,-0.0993, 139.727,-0.1687, 16.2656,-0.1134, 37.77,150., &
       0.3805, 825.41,-0.1073, 139.193,-0.1618, 19.0259,-0.1150, 38.84,170., &
       0.3988, 1359., -0.1016, 176.2,  -0.1836, 35.23,  -0.1198, 40.00,190./
     
! ExtendedFullFrame: parameters for 10 regions derived from merged Coma observations
  data efpar/                                                       &
       0.5284, 1385.9, 0.1524, 158.155, 0.1762, 19.5666, 0.1435, 31.05, 10., &
       0.5284, 1385.9, 0.1524, 158.155, 0.1762, 19.5666, 0.1435, 31.93, 30., &
       0.5284, 1385.9, 0.1524, 158.155, 0.1762, 19.5666, 0.1435, 32.84, 50., &
       0.5354, 1394.1, 0.1342, 200.864, 0.1629, 30.6517, 0.1627, 33.77, 70., &
       0.5320, 1499.4, 0.1311, 222.169, 0.1601, 35.5882, 0.1747, 34.73, 90., &
       0.5334, 1725.7, 0.1222, 260.906, 0.1489, 42.7871, 0.1926, 35.71,110., &
       0.5553, 1184.2, 0.1230, 190.263, 0.1669, 31.6242, 0.1550, 36.73,130., &
       0.5650, 955.74, 0.1335, 163.263, 0.1682, 25.0995, 0.1349, 37.77,150., &
       0.5649, 943.63, 0.1404, 161.558, 0.1691, 24.2733, 0.1295, 38.84,170., &
       0.5651, 963.51, 0.1460, 168.625, 0.1695, 22.5818, 0.1258, 40.00,190., &
       0.4101, 1356.3,-0.1121, 142.936,-0.1630, 19.8420,-0.1355, 31.05, 10., &
       0.4101, 1356.3,-0.1121, 142.936,-0.1630, 19.8420,-0.1355, 31.93, 30., &
       0.4101, 1356.3,-0.1121, 142.936,-0.1630, 19.8420,-0.1355, 32.84, 50., &
       0.4047, 1462.5,-0.0942, 193.856,-0.1428, 32.6295,-0.1629, 33.77, 70., &
       0.4015, 1484.8,-0.0849, 224.818,-0.1348, 38.8510,-0.1753, 34.73, 90., &
       0.4283, 2242.7,-0.1044, 259.620,-0.1286, 43.3512,-0.1936, 35.71,110., &
       0.3867, 987.81,-0.0955, 142.543,-0.1753, 15.0092,-0.1149, 36.73,130., &
       0.3812, 870.98,-0.0993, 139.727,-0.1687, 16.2656,-0.1134, 37.77,150., &
       0.3805, 825.41,-0.1073, 139.193,-0.1618, 19.0259,-0.1150, 38.84,170., &
       0.3823, 909.63,-0.1084, 156.678,-0.1599, 20.8893,-0.1210, 40.00,190./

! SmallWindow: parameters for focus region (Y9) derived from Mkn421 (ring), rev 0084
!              parameters for line 158 derived from Mkn421 (ring), rev 0084
!              then interpolated for Y6, Y7 and Y8
  data swpar/                                                           &
       0.5482,933.86, 0.1423,149.606, 0.1635,30.958, 0.1509,38.00,158., &
       0.5493,906.73, 0.1444,147.483, 0.1741,30.452, 0.1364,39.00,173., &
       0.5504,879.59, 0.1464,145.359, 0.1847,29.945, 0.1219,40.00,188., &
       0.3923,830.30,-0.1050,129.887,-0.1503,29.359,-0.1426,38.00,158., &
       0.3880,767.37,-0.1059,127.092,-0.1597,28.606,-0.1272,39.00,173., &
       0.3836,704.43,-0.1067,124.296,-0.1690,27.851,-0.1118,40.00,188./

! LargeWindow: parameters for focus region (Y9) derived from Mkn 205, rev 0075
! parameters derived with CTE version 1 May 2001
  data lwpar/                                                           &
       0.5334, 1725.7, 0.1222, 260.906, 0.1489, 42.7871, 0.1926, 35.71,110., &
       0.5553, 1184.2, 0.1230, 190.263, 0.1669, 31.6242, 0.1550, 36.73,130., &
       0.5650, 955.74, 0.1335, 163.263, 0.1682, 25.0995, 0.1349, 37.77,150., &
       0.5649, 943.63, 0.1404, 161.558, 0.1691, 24.2733, 0.1295, 38.84,170., &
       0.4913, 1613.,  0.1857, 175.7,  	0.2012, 30.95,   0.1223, 40.00,190., &
       0.4283, 2242.7,-0.1044, 259.620,-0.1286, 43.3512,-0.1936, 35.71,110., &
       0.3867, 987.81,-0.0955, 142.543,-0.1753, 15.0092,-0.1149, 36.73,130., &
       0.3812, 870.98,-0.0993, 139.727,-0.1687, 16.2656,-0.1134, 37.77,150., &
       0.3805, 825.41,-0.1073, 139.193,-0.1618, 19.0259,-0.1150, 38.84,170., &
       0.4336, 1706., -0.1301, 169.6,  -0.1817, 30.54,  -0.1227, 40.00,190./

! Timing mode: parameters derived from GX13+1, rev 0057
! parameters derived with CTE version Apr 2001
  data tipar/                                                           &
       0.5805,6118.,  0.2479,965.9,   0.0538,135.2,  0.1100,80.00,190., &
       0.3078,4585., -0.1425,968.2,  -0.0494,122.9, -0.1150,80.00,190./

! Burst mode: parameters derived from XTEJ1650, rev 0323
!             (consistent with Crab, rev 0234)
  data bupar/                                                           &
       0.6194,2173.,  0.1194,473.0,   0.1094,63.87,  0.1475,46.00,190., &
       0.3383,1963., -0.0982,420.9,  -0.0947,60.18, -0.1442,46.00,190./
      

  irun = irun+1

! if (ccdmode.eq.'FF') then
  if (calMode == 0) then
    iy1 = int(line/20.) + 1
    if (iy1.ge.maxff) iy1 = maxff - 1
    if (iy1.lt.1) iy1 = 1
    iy2 = iy1 + 1
    rly1 = 20.0*real(iy1) - 10.0
    rly2 = rly1 + 20.0
    do ipar = 1,npar
      par(ipar) = ffpar(ipar,iy2,pattyp) -		   &
 	    (ffpar(ipar,iy2,pattyp)-ffpar(ipar,iy1,pattyp))*   &
 	    (rly2-line)/20.0
    enddo
! elseif (ccdmode.eq.'EF') then
  elseif (calMode == 1) then
    iy1 = int(line/20.) + 1
    if (iy1.ge.maxef) iy1 = maxef - 1
    if (iy1.lt.1) iy1 = 1
    iy2 = iy1 + 1
    rly1 = 20.0*real(iy1) - 10.0
    rly2 = rly1 + 20.0
    do ipar = 1,npar
      par(ipar) = efpar(ipar,iy2,pattyp) -		   &
 	    (efpar(ipar,iy2,pattyp)-efpar(ipar,iy1,pattyp))*   &
 	    (rly2-line)/20.0
    enddo
! elseif (ccdmode.eq.'SW') then
  elseif (calMode == 3) then
    iy1 = int(line/20.) - 6
    if (iy1.ge.maxsw) iy1 = maxsw - 1
    if (iy1.lt.1) iy1 = 1
    iy2 = iy1 + 1
    rly1 = 20.0*real(iy1) + 130.0
    rly2 = rly1 + 20.0
    do ipar = 1,npar
      par(ipar) = swpar(ipar,iy2,pattyp) -		   &
 	    (swpar(ipar,iy2,pattyp)-swpar(ipar,iy1,pattyp))*   &
 	    (rly2-line)/20.0
    enddo
! elseif (ccdmode.eq.'LW') then
  elseif (calMode == 2) then
    iy1 = int(line/20.) - 4
    if (iy1.ge.maxlw) iy1 = maxlw - 1
    if (iy1.lt.1) iy1 = 1
    iy2 = iy1 + 1
    rly1 = 20.0*real(iy1) + 90.0
    rly2 = rly1 + 20.0
    do ipar = 1,npar
      par(ipar) = lwpar(ipar,iy2,pattyp) -		   &
     	(lwpar(ipar,iy2,pattyp)-lwpar(ipar,iy1,pattyp))*   &
     	(rly2-line)/20.0
    enddo
! elseif (ccdmode.eq.'TI') then
  elseif (calMode == 4) then
    iy1 = 1
    do ipar = 1,npar
      par(ipar) = tipar(ipar,iy1,pattyp)
    enddo
! elseif (ccdmode.eq.'BU') then
  elseif (calMode == 5) then
    iy1 = 1
    do ipar = 1,npar
      par(ipar) = bupar(ipar,iy1,pattyp)
    enddo
  else
    do ipar = 1,7,2
      par(ipar) = 0.0
    enddo
    do ipar = 2,6,2
      par(ipar) = 1.0
    enddo
  endif
! par(8) = real(dthresh)/1.0014**(191.0-line)
  ndthresh = int(par(8))
  if (irun .le. 1) then
    call message(AppMsg,Verbose,' ')
    write(errstr,'(1x,a,t35,f9.4,i8)') 'Effective threshold: ',par(8),dthresh
    call message(AppMsg,Verbose)
  endif

! single-event pattern
  if (pattyp.eq.1) then
    do ibin = 1, ndthresh-1
      patfraction(ibin) = 1.0
    enddo
    do ibin = ndthresh, maxbin
      rbin = 2.0*(real(ibin)-par(8))
!     if (calMode < 5) then
       	patfraction(ibin) = min(1.0,		    &
 	                          par(1)+par(3)/(1.0+(rbin/par(2))**2)   &
 		                        +par(5)/(1.0+(rbin/par(4))**2)   &
 		                        +par(7)/(1.0+(rbin/par(6))**2))
!     else
!	      patfraction(ibin) = min(1.0,0.6915-0.125/0.69897*log10(real(ibin)/1000.))     
!     endif
    enddo
! double-event pattern
  elseif (pattyp.eq.2) then
    do ibin = 1, ndthresh-1
      patfraction(ibin) = 0.0
    enddo
    do ibin = ndthresh, maxbin
      rbin = 2.0*(real(ibin)-par(8))
!     if (calMode < 5) then
 	      patfraction(ibin) = max(0.0,		    &
 	                          par(1)+par(3)/(1.0+(rbin/par(2))**2)   &
 		                        +par(5)/(1.0+(rbin/par(4))**2)   &
 		                        +par(7)/(1.0+(rbin/par(6))**2))
!     else
!	      patfraction(ibin) = max(0.0,0.285+0.108/0.69897*log10(real(ibin)/1000.))      
!     endif
    enddo
  endif

end subroutine patmod

! ==========================================================================
! returns the pattern distribution model
! ==========================================================================
subroutine patmod_mos(patfraction, maxbin, energy, pattyp)

  implicit none
  integer          ,intent(in)    :: pattyp, maxbin
  real(kind=single),intent(out)   :: patfraction(1:maxbin),energy(maxbin)

  integer,parameter               :: nbin = 800
  integer                         :: irun = 0
  real(kind=single)               :: e(1:nbin),s(1:nbin),d(1:nbin), &
                                     t(1:nbin),q(1:nbin)

  if (irun == 0) then

  E(   1: 160) = (/    0.0,   15.0,   30.0,   45.0,   60.0,   75.0,   90.0,  105.0,&
                     120.0,  135.0,  150.0,  165.0,  180.0,  195.0,  210.0,  225.0,&
                     240.0,  255.0,  270.0,  285.0,  300.0,  315.0,  330.0,  345.0,&
                     360.0,  375.0,  390.0,  405.0,  420.0,  435.0,  450.0,  465.0,&
                     480.0,  495.0,  510.0,  525.0,  540.0,  555.0,  570.0,  585.0,&
                     600.0,  615.0,  630.0,  645.0,  660.0,  675.0,  690.0,  705.0,&
                     720.0,  735.0,  750.0,  765.0,  780.0,  795.0,  810.0,  825.0,&
                     840.0,  855.0,  870.0,  885.0,  900.0,  915.0,  930.0,  945.0,&
                     960.0,  975.0,  990.0, 1005.0, 1020.0, 1035.0, 1050.0, 1065.0,&
                    1080.0, 1095.0, 1110.0, 1125.0, 1140.0, 1155.0, 1170.0, 1185.0,&
                    1200.0, 1215.0, 1230.0, 1245.0, 1260.0, 1275.0, 1290.0, 1305.0,&
                    1320.0, 1335.0, 1350.0, 1365.0, 1380.0, 1395.0, 1410.0, 1425.0,&
                    1440.0, 1455.0, 1470.0, 1485.0, 1500.0, 1515.0, 1530.0, 1545.0,&
                    1560.0, 1575.0, 1590.0, 1605.0, 1620.0, 1635.0, 1650.0, 1665.0,&
                    1680.0, 1695.0, 1710.0, 1725.0, 1740.0, 1755.0, 1770.0, 1785.0,&
                    1800.0, 1815.0, 1830.0, 1845.0, 1860.0, 1875.0, 1890.0, 1905.0,&
                    1920.0, 1935.0, 1950.0, 1965.0, 1980.0, 1995.0, 2010.0, 2025.0,&
                    2040.0, 2055.0, 2070.0, 2085.0, 2100.0, 2115.0, 2130.0, 2145.0,&
                    2160.0, 2175.0, 2190.0, 2205.0, 2220.0, 2235.0, 2250.0, 2265.0,&
                    2280.0, 2295.0, 2310.0, 2325.0, 2340.0, 2355.0, 2370.0, 2385.0 /)
  E( 161: 320) = (/ 2400.0, 2415.0, 2430.0, 2445.0, 2460.0, 2475.0, 2490.0, 2505.0,&
                    2520.0, 2535.0, 2550.0, 2565.0, 2580.0, 2595.0, 2610.0, 2625.0,&
                    2640.0, 2655.0, 2670.0, 2685.0, 2700.0, 2715.0, 2730.0, 2745.0,&
                    2760.0, 2775.0, 2790.0, 2805.0, 2820.0, 2835.0, 2850.0, 2865.0,&
                    2880.0, 2895.0, 2910.0, 2925.0, 2940.0, 2955.0, 2970.0, 2985.0,&
                    3000.0, 3015.0, 3030.0, 3045.0, 3060.0, 3075.0, 3090.0, 3105.0,&
                    3120.0, 3135.0, 3150.0, 3165.0, 3180.0, 3195.0, 3210.0, 3225.0,&
                    3240.0, 3255.0, 3270.0, 3285.0, 3300.0, 3315.0, 3330.0, 3345.0,&
                    3360.0, 3375.0, 3390.0, 3405.0, 3420.0, 3435.0, 3450.0, 3465.0,&
                    3480.0, 3495.0, 3510.0, 3525.0, 3540.0, 3555.0, 3570.0, 3585.0,&
                    3600.0, 3615.0, 3630.0, 3645.0, 3660.0, 3675.0, 3690.0, 3705.0,&
                    3720.0, 3735.0, 3750.0, 3765.0, 3780.0, 3795.0, 3810.0, 3825.0,&
                    3840.0, 3855.0, 3870.0, 3885.0, 3900.0, 3915.0, 3930.0, 3945.0,&
                    3960.0, 3975.0, 3990.0, 4005.0, 4020.0, 4035.0, 4050.0, 4065.0,&
                    4080.0, 4095.0, 4110.0, 4125.0, 4140.0, 4155.0, 4170.0, 4185.0,&
                    4200.0, 4215.0, 4230.0, 4245.0, 4260.0, 4275.0, 4290.0, 4305.0,&
                    4320.0, 4335.0, 4350.0, 4365.0, 4380.0, 4395.0, 4410.0, 4425.0,&
                    4440.0, 4455.0, 4470.0, 4485.0, 4500.0, 4515.0, 4530.0, 4545.0,&
                    4560.0, 4575.0, 4590.0, 4605.0, 4620.0, 4635.0, 4650.0, 4665.0,&
                    4680.0, 4695.0, 4710.0, 4725.0, 4740.0, 4755.0, 4770.0, 4785.0 /)
  E( 321: 480) = (/ 4800.0, 4815.0, 4830.0, 4845.0, 4860.0, 4875.0, 4890.0, 4905.0,&
                    4920.0, 4935.0, 4950.0, 4965.0, 4980.0, 4995.0, 5010.0, 5025.0,&
                    5040.0, 5055.0, 5070.0, 5085.0, 5100.0, 5115.0, 5130.0, 5145.0,&
                    5160.0, 5175.0, 5190.0, 5205.0, 5220.0, 5235.0, 5250.0, 5265.0,&
                    5280.0, 5295.0, 5310.0, 5325.0, 5340.0, 5355.0, 5370.0, 5385.0,&
                    5400.0, 5415.0, 5430.0, 5445.0, 5460.0, 5475.0, 5490.0, 5505.0,&
                    5520.0, 5535.0, 5550.0, 5565.0, 5580.0, 5595.0, 5610.0, 5625.0,&
                    5640.0, 5655.0, 5670.0, 5685.0, 5700.0, 5715.0, 5730.0, 5745.0,&
                    5760.0, 5775.0, 5790.0, 5805.0, 5820.0, 5835.0, 5850.0, 5865.0,&
                    5880.0, 5895.0, 5910.0, 5925.0, 5940.0, 5955.0, 5970.0, 5985.0,&
                    6000.0, 6015.0, 6030.0, 6045.0, 6060.0, 6075.0, 6090.0, 6105.0,&
                    6120.0, 6135.0, 6150.0, 6165.0, 6180.0, 6195.0, 6210.0, 6225.0,&
                    6240.0, 6255.0, 6270.0, 6285.0, 6300.0, 6315.0, 6330.0, 6345.0,&
                    6360.0, 6375.0, 6390.0, 6405.0, 6420.0, 6435.0, 6450.0, 6465.0,&
                    6480.0, 6495.0, 6510.0, 6525.0, 6540.0, 6555.0, 6570.0, 6585.0,&
                    6600.0, 6615.0, 6630.0, 6645.0, 6660.0, 6675.0, 6690.0, 6705.0,&
                    6720.0, 6735.0, 6750.0, 6765.0, 6780.0, 6795.0, 6810.0, 6825.0,&
                    6840.0, 6855.0, 6870.0, 6885.0, 6900.0, 6915.0, 6930.0, 6945.0,&
                    6960.0, 6975.0, 6990.0, 7005.0, 7020.0, 7035.0, 7050.0, 7065.0,&
                    7080.0, 7095.0, 7110.0, 7125.0, 7140.0, 7155.0, 7170.0, 7185.0 /)
  E( 481: 640) = (/ 7200.0, 7215.0, 7230.0, 7245.0, 7260.0, 7275.0, 7290.0, 7305.0,&
                    7320.0, 7335.0, 7350.0, 7365.0, 7380.0, 7395.0, 7410.0, 7425.0,&
                    7440.0, 7455.0, 7470.0, 7485.0, 7500.0, 7515.0, 7530.0, 7545.0,&
                    7560.0, 7575.0, 7590.0, 7605.0, 7620.0, 7635.0, 7650.0, 7665.0,&
                    7680.0, 7695.0, 7710.0, 7725.0, 7740.0, 7755.0, 7770.0, 7785.0,&
                    7800.0, 7815.0, 7830.0, 7845.0, 7860.0, 7875.0, 7890.0, 7905.0,&
                    7920.0, 7935.0, 7950.0, 7965.0, 7980.0, 7995.0, 8010.0, 8025.0,&
                    8040.0, 8055.0, 8070.0, 8085.0, 8100.0, 8115.0, 8130.0, 8145.0,&
                    8160.0, 8175.0, 8190.0, 8205.0, 8220.0, 8235.0, 8250.0, 8265.0,&
                    8280.0, 8295.0, 8310.0, 8325.0, 8340.0, 8355.0, 8370.0, 8385.0,&
                    8400.0, 8415.0, 8430.0, 8445.0, 8460.0, 8475.0, 8490.0, 8505.0,&
                    8520.0, 8535.0, 8550.0, 8565.0, 8580.0, 8595.0, 8610.0, 8625.0,&
                    8640.0, 8655.0, 8670.0, 8685.0, 8700.0, 8715.0, 8730.0, 8745.0,&
                    8760.0, 8775.0, 8790.0, 8805.0, 8820.0, 8835.0, 8850.0, 8865.0,&
                    8880.0, 8895.0, 8910.0, 8925.0, 8940.0, 8955.0, 8970.0, 8985.0,&
                    9000.0, 9015.0, 9030.0, 9045.0, 9060.0, 9075.0, 9090.0, 9105.0,&
                    9120.0, 9135.0, 9150.0, 9165.0, 9180.0, 9195.0, 9210.0, 9225.0,&
                    9240.0, 9255.0, 9270.0, 9285.0, 9300.0, 9315.0, 9330.0, 9345.0,&
                    9360.0, 9375.0, 9390.0, 9405.0, 9420.0, 9435.0, 9450.0, 9465.0,&
                    9480.0, 9495.0, 9510.0, 9525.0, 9540.0, 9555.0, 9570.0, 9585.0 /)
  E( 641: 800) = (/ 9600.0, 9615.0, 9630.0, 9645.0, 9660.0, 9675.0, 9690.0, 9705.0,&
                    9720.0, 9735.0, 9750.0, 9765.0, 9780.0, 9795.0, 9810.0, 9825.0,&
                    9840.0, 9855.0, 9870.0, 9885.0, 9900.0, 9915.0, 9930.0, 9945.0,&
                    9960.0, 9975.0, 9990.0,10005.0,10020.0,10035.0,10050.0,10065.0,&
                   10080.0,10095.0,10110.0,10125.0,10140.0,10155.0,10170.0,10185.0,&
                   10200.0,10215.0,10230.0,10245.0,10260.0,10275.0,10290.0,10305.0,&
                   10320.0,10335.0,10350.0,10365.0,10380.0,10395.0,10410.0,10425.0,&
                   10440.0,10455.0,10470.0,10485.0,10500.0,10515.0,10530.0,10545.0,&
                   10560.0,10575.0,10590.0,10605.0,10620.0,10635.0,10650.0,10665.0,&
                   10680.0,10695.0,10710.0,10725.0,10740.0,10755.0,10770.0,10785.0,&
                   10800.0,10815.0,10830.0,10845.0,10860.0,10875.0,10890.0,10905.0,&
                   10920.0,10935.0,10950.0,10965.0,10980.0,10995.0,11010.0,11025.0,&
                   11040.0,11055.0,11070.0,11085.0,11100.0,11115.0,11130.0,11145.0,&
                   11160.0,11175.0,11190.0,11205.0,11220.0,11235.0,11250.0,11265.0,&
                   11280.0,11295.0,11310.0,11325.0,11340.0,11355.0,11370.0,11385.0,&
                   11400.0,11415.0,11430.0,11445.0,11460.0,11475.0,11490.0,11505.0,&
                   11520.0,11535.0,11550.0,11565.0,11580.0,11595.0,11610.0,11625.0,&
                   11640.0,11655.0,11670.0,11685.0,11700.0,11715.0,11730.0,11745.0,&
                   11760.0,11775.0,11790.0,11805.0,11820.0,11835.0,11850.0,11865.0,&
                   11880.0,11895.0,11910.0,11925.0,11940.0,11955.0,11970.0,11985.0 /)
  S(   1: 160) = (/0.00000,0.00000,0.00000,0.00000,0.00000,1.00000,1.00000,1.00000,&
                   1.00000,1.00000,1.00000,0.99155,0.95250,0.95029,0.94271,0.93423,&
                   0.92657,0.92039,0.91583,0.91271,0.91037,0.90811,0.90551,0.90244,&
                   0.89887,0.89549,0.89197,0.88881,0.88684,0.88634,0.88657,0.88669,&
                   0.88626,0.88511,0.88315,0.88036,0.87725,0.87473,0.87309,0.87188,&
                   0.87076,0.86965,0.86848,0.86709,0.86533,0.86327,0.86122,0.85941,&
                   0.85788,0.85658,0.85545,0.85445,0.85359,0.85284,0.85221,0.85171,&
                   0.85137,0.85123,0.85129,0.85140,0.85135,0.85102,0.85033,0.84926,&
                   0.84783,0.84615,0.84432,0.84246,0.84062,0.83881,0.83700,0.83515,&
                   0.83323,0.83116,0.82884,0.82615,0.82300,0.81946,0.81585,0.81259,&
                   0.80993,0.80789,0.80630,0.80496,0.80364,0.80218,0.80049,0.79855,&
                   0.79640,0.79408,0.79166,0.78917,0.78665,0.78411,0.78152,0.77887,&
                   0.77613,0.77333,0.77051,0.76775,0.76514,0.76275,0.76065,0.75880,&
                   0.75712,0.75551,0.75394,0.75249,0.75126,0.75036,0.74990,0.75010,&
                   0.75129,0.75366,0.75678,0.75965,0.76127,0.76159,0.76149,0.76218,&
                   0.76438,0.76818,0.77339,0.77975,0.78699,0.79421,0.80013,0.80401,&
                   0.80600,0.80670,0.80671,0.80645,0.80608,0.80568,0.80525,0.80481,&
                   0.80438,0.80400,0.80369,0.80346,0.80333,0.80328,0.80330,0.80335,&
                   0.80339,0.80336,0.80323,0.80300,0.80265,0.80221,0.80173,0.80122,&
                   0.80073,0.80027,0.79987,0.79950,0.79916,0.79882,0.79847,0.79810 /)
  S( 161: 320) = (/0.79769,0.79724,0.79674,0.79619,0.79559,0.79494,0.79425,0.79353,&
                   0.79281,0.79209,0.79140,0.79074,0.79012,0.78954,0.78900,0.78849,&
                   0.78800,0.78752,0.78705,0.78658,0.78611,0.78565,0.78518,0.78470,&
                   0.78422,0.78372,0.78320,0.78265,0.78208,0.78149,0.78090,0.78032,&
                   0.77980,0.77934,0.77896,0.77860,0.77813,0.77739,0.77620,0.77449,&
                   0.77235,0.77007,0.76804,0.76652,0.76550,0.76477,0.76411,0.76337,&
                   0.76251,0.76156,0.76055,0.75952,0.75849,0.75747,0.75646,0.75547,&
                   0.75450,0.75355,0.75262,0.75171,0.75081,0.74993,0.74905,0.74818,&
                   0.74730,0.74643,0.74554,0.74464,0.74372,0.74279,0.74183,0.74087,&
                   0.73990,0.73893,0.73796,0.73699,0.73603,0.73506,0.73409,0.73313,&
                   0.73216,0.73119,0.73021,0.72924,0.72827,0.72730,0.72634,0.72538,&
                   0.72443,0.72349,0.72255,0.72163,0.72072,0.71982,0.71894,0.71808,&
                   0.71723,0.71639,0.71557,0.71475,0.71395,0.71314,0.71233,0.71150,&
                   0.71065,0.70977,0.70886,0.70793,0.70697,0.70600,0.70502,0.70404,&
                   0.70306,0.70209,0.70112,0.70016,0.69919,0.69822,0.69724,0.69624,&
                   0.69523,0.69420,0.69315,0.69207,0.69097,0.68985,0.68870,0.68754,&
                   0.68637,0.68519,0.68401,0.68284,0.68168,0.68054,0.67943,0.67835,&
                   0.67730,0.67629,0.67533,0.67440,0.67351,0.67265,0.67182,0.67102,&
                   0.67023,0.66946,0.66870,0.66796,0.66722,0.66649,0.66578,0.66507,&
                   0.66436,0.66367,0.66298,0.66230,0.66162,0.66095,0.66029,0.65964 /)
  S( 321: 480) = (/0.65899,0.65834,0.65771,0.65708,0.65646,0.65584,0.65523,0.65463,&
                   0.65403,0.65343,0.65283,0.65223,0.65162,0.65101,0.65039,0.64977,&
                   0.64914,0.64852,0.64790,0.64728,0.64667,0.64605,0.64545,0.64484,&
                   0.64424,0.64363,0.64304,0.64244,0.64185,0.64126,0.64068,0.64009,&
                   0.63952,0.63894,0.63837,0.63780,0.63723,0.63667,0.63612,0.63556,&
                   0.63502,0.63448,0.63394,0.63342,0.63290,0.63240,0.63191,0.63144,&
                   0.63098,0.63053,0.63011,0.62970,0.62931,0.62894,0.62860,0.62828,&
                   0.62800,0.62776,0.62756,0.62740,0.62731,0.62729,0.62734,0.62747,&
                   0.62770,0.62804,0.62849,0.62905,0.62973,0.63052,0.63141,0.63235,&
                   0.63329,0.63414,0.63476,0.63501,0.63474,0.63383,0.63225,0.63008,&
                   0.62752,0.62484,0.62231,0.62019,0.61861,0.61761,0.61709,0.61687,&
                   0.61680,0.61677,0.61671,0.61660,0.61645,0.61626,0.61605,0.61583,&
                   0.61560,0.61537,0.61515,0.61492,0.61470,0.61447,0.61425,0.61403,&
                   0.61381,0.61358,0.61336,0.61313,0.61291,0.61268,0.61245,0.61223,&
                   0.61200,0.61178,0.61155,0.61133,0.61112,0.61090,0.61068,0.61047,&
                   0.61026,0.61005,0.60984,0.60963,0.60943,0.60922,0.60902,0.60882,&
                   0.60861,0.60841,0.60820,0.60800,0.60779,0.60759,0.60738,0.60717,&
                   0.60696,0.60676,0.60655,0.60635,0.60615,0.60595,0.60575,0.60555,&
                   0.60535,0.60514,0.60492,0.60470,0.60447,0.60423,0.60399,0.60374,&
                   0.60350,0.60326,0.60303,0.60281,0.60260,0.60240,0.60222,0.60205 /)
  S( 481: 640) = (/0.60189,0.60174,0.60161,0.60149,0.60137,0.60127,0.60117,0.60108,&
                   0.60099,0.60091,0.60082,0.60072,0.60061,0.60049,0.60035,0.60018,&
                   0.60000,0.59979,0.59957,0.59933,0.59908,0.59882,0.59856,0.59829,&
                   0.59801,0.59773,0.59744,0.59715,0.59685,0.59655,0.59625,0.59595,&
                   0.59565,0.59535,0.59507,0.59479,0.59453,0.59427,0.59403,0.59380,&
                   0.59359,0.59340,0.59322,0.59307,0.59293,0.59282,0.59273,0.59266,&
                   0.59261,0.59258,0.59257,0.59257,0.59259,0.59262,0.59266,0.59272,&
                   0.59279,0.59287,0.59298,0.59309,0.59323,0.59337,0.59351,0.59366,&
                   0.59381,0.59395,0.59407,0.59417,0.59425,0.59430,0.59432,0.59430,&
                   0.59426,0.59418,0.59407,0.59394,0.59377,0.59359,0.59338,0.59315,&
                   0.59291,0.59264,0.59237,0.59208,0.59179,0.59148,0.59117,0.59085,&
                   0.59052,0.59018,0.58984,0.58950,0.58915,0.58880,0.58845,0.58810,&
                   0.58775,0.58740,0.58706,0.58671,0.58637,0.58604,0.58570,0.58538,&
                   0.58506,0.58475,0.58445,0.58415,0.58386,0.58358,0.58332,0.58306,&
                   0.58282,0.58259,0.58238,0.58219,0.58202,0.58188,0.58177,0.58170,&
                   0.58166,0.58167,0.58171,0.58179,0.58191,0.58206,0.58224,0.58246,&
                   0.58270,0.58297,0.58326,0.58356,0.58389,0.58422,0.58457,0.58492,&
                   0.58528,0.58564,0.58600,0.58635,0.58670,0.58704,0.58736,0.58767,&
                   0.58797,0.58824,0.58849,0.58872,0.58892,0.58909,0.58923,0.58934,&
                   0.58942,0.58946,0.58948,0.58946,0.58941,0.58933,0.58923,0.58910 /)
  S( 641: 800) = (/0.58895,0.58878,0.58859,0.58838,0.58816,0.58792,0.58767,0.58741,&
                   0.58713,0.58685,0.58656,0.58627,0.58596,0.58565,0.58534,0.58502,&
                   0.58470,0.58437,0.58404,0.58370,0.58337,0.58303,0.58270,0.58237,&
                   0.58206,0.58177,0.58149,0.58124,0.58101,0.58080,0.58062,0.58044,&
                   0.58028,0.58013,0.57999,0.57985,0.57971,0.57958,0.57944,0.57931,&
                   0.57917,0.57903,0.57889,0.57875,0.57861,0.57847,0.57833,0.57819,&
                   0.57805,0.57792,0.57778,0.57765,0.57751,0.57738,0.57724,0.57710,&
                   0.57695,0.57681,0.57666,0.57651,0.57636,0.57621,0.57606,0.57591,&
                   0.57575,0.57559,0.57543,0.57527,0.57511,0.57495,0.57479,0.57462,&
                   0.57445,0.57429,0.57412,0.57395,0.57378,0.57362,0.57345,0.57328,&
                   0.57312,0.57295,0.57279,0.57263,0.57247,0.57231,0.57215,0.57199,&
                   0.57183,0.57167,0.57151,0.57135,0.57119,0.57102,0.57086,0.57070,&
                   0.57053,0.57037,0.57021,0.57005,0.56989,0.56973,0.56957,0.56941,&
                   0.56926,0.56910,0.56894,0.56879,0.56864,0.56848,0.56833,0.56817,&
                   0.56802,0.56786,0.56771,0.56755,0.56740,0.56725,0.56709,0.56694,&
                   0.56679,0.56664,0.56649,0.56635,0.56620,0.56607,0.56593,0.56581,&
                   0.56569,0.56558,0.56547,0.56538,0.56529,0.56522,0.56515,0.56510,&
                   0.56507,0.56506,0.56508,0.56515,0.56529,0.56553,0.56589,0.56643,&
                   0.56718,0.56820,0.56952,0.57118,0.57319,0.57552,0.57814,0.58098,&
                   0.58393,0.58689,0.58975,0.59237,0.59464,0.59643,0.59761,0.59805 /)
  D(   1: 160) = (/0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,&
                   0.00000,0.00000,0.00000,0.00845,0.04750,0.04971,0.05729,0.06577,&
                   0.07343,0.07959,0.08409,0.08719,0.08952,0.09178,0.09437,0.09733,&
                   0.10064,0.10418,0.10769,0.11081,0.11272,0.11315,0.11285,0.11267,&
                   0.11304,0.11415,0.11609,0.11886,0.12194,0.12445,0.12608,0.12725,&
                   0.12832,0.12936,0.13044,0.13172,0.13335,0.13525,0.13714,0.13878,&
                   0.14014,0.14129,0.14228,0.14314,0.14387,0.14449,0.14498,0.14536,&
                   0.14558,0.14561,0.14545,0.14526,0.14522,0.14546,0.14603,0.14695,&
                   0.14820,0.14970,0.15133,0.15299,0.15463,0.15625,0.15785,0.15945,&
                   0.16109,0.16282,0.16473,0.16695,0.16959,0.17258,0.17564,0.17836,&
                   0.18046,0.18192,0.18289,0.18354,0.18406,0.18459,0.18522,0.18595,&
                   0.18676,0.18762,0.18847,0.18929,0.19008,0.19083,0.19157,0.19233,&
                   0.19314,0.19401,0.19489,0.19569,0.19632,0.19667,0.19666,0.19633,&
                   0.19577,0.19507,0.19427,0.19333,0.19220,0.19086,0.18930,0.18748,&
                   0.18523,0.18247,0.17942,0.17672,0.17504,0.17461,0.17516,0.17625,&
                   0.17768,0.17936,0.18123,0.18298,0.18402,0.18415,0.18384,0.18367,&
                   0.18384,0.18425,0.18472,0.18515,0.18552,0.18586,0.18619,0.18650,&
                   0.18679,0.18704,0.18723,0.18735,0.18738,0.18735,0.18726,0.18715,&
                   0.18705,0.18702,0.18709,0.18726,0.18754,0.18790,0.18831,0.18872,&
                   0.18910,0.18943,0.18968,0.18988,0.19004,0.19020,0.19038,0.19061 /)
  D( 161: 320) = (/0.19091,0.19127,0.19171,0.19221,0.19276,0.19336,0.19397,0.19460,&
                   0.19523,0.19583,0.19641,0.19695,0.19744,0.19790,0.19832,0.19871,&
                   0.19908,0.19943,0.19976,0.20008,0.20039,0.20068,0.20097,0.20125,&
                   0.20153,0.20183,0.20215,0.20249,0.20286,0.20326,0.20367,0.20409,&
                   0.20450,0.20487,0.20520,0.20553,0.20591,0.20648,0.20736,0.20861,&
                   0.21015,0.21178,0.21323,0.21434,0.21512,0.21570,0.21626,0.21686,&
                   0.21752,0.21822,0.21892,0.21963,0.22031,0.22098,0.22162,0.22224,&
                   0.22283,0.22340,0.22395,0.22447,0.22497,0.22546,0.22594,0.22641,&
                   0.22687,0.22734,0.22780,0.22828,0.22877,0.22927,0.22978,0.23029,&
                   0.23080,0.23131,0.23180,0.23228,0.23275,0.23321,0.23366,0.23411,&
                   0.23456,0.23500,0.23544,0.23588,0.23630,0.23673,0.23714,0.23755,&
                   0.23795,0.23834,0.23872,0.23908,0.23944,0.23977,0.24009,0.24039,&
                   0.24068,0.24095,0.24121,0.24146,0.24170,0.24193,0.24217,0.24243,&
                   0.24270,0.24301,0.24334,0.24371,0.24409,0.24449,0.24490,0.24531,&
                   0.24570,0.24609,0.24648,0.24686,0.24723,0.24762,0.24801,0.24841,&
                   0.24883,0.24928,0.24975,0.25024,0.25077,0.25132,0.25189,0.25248,&
                   0.25310,0.25372,0.25436,0.25499,0.25562,0.25623,0.25683,0.25741,&
                   0.25796,0.25848,0.25897,0.25943,0.25986,0.26026,0.26064,0.26101,&
                   0.26136,0.26170,0.26202,0.26234,0.26265,0.26295,0.26325,0.26354,&
                   0.26382,0.26410,0.26438,0.26465,0.26491,0.26517,0.26543,0.26569 /)
  D( 321: 480) = (/0.26594,0.26619,0.26643,0.26667,0.26692,0.26715,0.26739,0.26763,&
                   0.26787,0.26810,0.26835,0.26860,0.26885,0.26912,0.26938,0.26966,&
                   0.26994,0.27022,0.27050,0.27079,0.27107,0.27136,0.27164,0.27193,&
                   0.27222,0.27250,0.27279,0.27308,0.27336,0.27365,0.27393,0.27421,&
                   0.27449,0.27477,0.27505,0.27533,0.27560,0.27588,0.27615,0.27641,&
                   0.27668,0.27694,0.27719,0.27743,0.27767,0.27790,0.27811,0.27832,&
                   0.27851,0.27869,0.27885,0.27900,0.27912,0.27923,0.27931,0.27936,&
                   0.27938,0.27937,0.27931,0.27920,0.27904,0.27882,0.27852,0.27815,&
                   0.27769,0.27716,0.27654,0.27585,0.27509,0.27428,0.27344,0.27260,&
                   0.27180,0.27110,0.27058,0.27031,0.27039,0.27086,0.27177,0.27306,&
                   0.27464,0.27637,0.27812,0.27975,0.28115,0.28227,0.28312,0.28372,&
                   0.28415,0.28447,0.28472,0.28492,0.28510,0.28528,0.28544,0.28560,&
                   0.28576,0.28592,0.28607,0.28623,0.28638,0.28653,0.28668,0.28683,&
                   0.28697,0.28712,0.28726,0.28740,0.28754,0.28767,0.28780,0.28793,&
                   0.28805,0.28817,0.28828,0.28838,0.28849,0.28858,0.28867,0.28876,&
                   0.28884,0.28892,0.28900,0.28907,0.28914,0.28921,0.28927,0.28934,&
                   0.28940,0.28947,0.28954,0.28960,0.28967,0.28974,0.28981,0.28989,&
                   0.28996,0.29004,0.29012,0.29019,0.29027,0.29035,0.29042,0.29050,&
                   0.29059,0.29068,0.29078,0.29088,0.29099,0.29112,0.29124,0.29137,&
                   0.29151,0.29164,0.29177,0.29190,0.29202,0.29213,0.29224,0.29234 /)
  D( 481: 640) = (/0.29243,0.29252,0.29260,0.29267,0.29273,0.29279,0.29285,0.29290,&
                   0.29295,0.29299,0.29304,0.29309,0.29315,0.29322,0.29330,0.29340,&
                   0.29351,0.29364,0.29379,0.29395,0.29413,0.29432,0.29452,0.29473,&
                   0.29495,0.29516,0.29538,0.29560,0.29582,0.29603,0.29625,0.29645,&
                   0.29665,0.29684,0.29702,0.29718,0.29734,0.29749,0.29762,0.29774,&
                   0.29784,0.29794,0.29801,0.29808,0.29812,0.29816,0.29817,0.29818,&
                   0.29818,0.29817,0.29815,0.29814,0.29812,0.29811,0.29810,0.29810,&
                   0.29810,0.29809,0.29809,0.29809,0.29808,0.29808,0.29807,0.29807,&
                   0.29807,0.29808,0.29809,0.29811,0.29815,0.29820,0.29826,0.29833,&
                   0.29842,0.29853,0.29865,0.29877,0.29892,0.29907,0.29922,0.29939,&
                   0.29957,0.29975,0.29993,0.30012,0.30032,0.30052,0.30072,0.30093,&
                   0.30114,0.30135,0.30156,0.30178,0.30199,0.30221,0.30242,0.30264,&
                   0.30285,0.30307,0.30328,0.30349,0.30370,0.30391,0.30411,0.30431,&
                   0.30451,0.30470,0.30489,0.30508,0.30526,0.30544,0.30562,0.30579,&
                   0.30595,0.30610,0.30625,0.30639,0.30651,0.30661,0.30670,0.30677,&
                   0.30681,0.30683,0.30683,0.30680,0.30675,0.30668,0.30658,0.30647,&
                   0.30634,0.30619,0.30603,0.30586,0.30568,0.30550,0.30530,0.30511,&
                   0.30491,0.30471,0.30451,0.30432,0.30413,0.30395,0.30377,0.30360,&
                   0.30345,0.30330,0.30318,0.30306,0.30296,0.30288,0.30282,0.30278,&
                   0.30276,0.30276,0.30277,0.30281,0.30286,0.30293,0.30302,0.30312 /)
  D( 641: 800) = (/0.30324,0.30337,0.30351,0.30366,0.30383,0.30400,0.30418,0.30437,&
                   0.30457,0.30477,0.30498,0.30520,0.30542,0.30565,0.30589,0.30613,&
                   0.30638,0.30663,0.30689,0.30715,0.30742,0.30769,0.30796,0.30823,&
                   0.30849,0.30874,0.30897,0.30920,0.30941,0.30961,0.30980,0.30997,&
                   0.31014,0.31030,0.31045,0.31060,0.31075,0.31089,0.31104,0.31118,&
                   0.31133,0.31147,0.31162,0.31177,0.31191,0.31206,0.31220,0.31234,&
                   0.31248,0.31261,0.31274,0.31287,0.31299,0.31311,0.31322,0.31333,&
                   0.31344,0.31354,0.31364,0.31373,0.31382,0.31391,0.31399,0.31406,&
                   0.31414,0.31420,0.31427,0.31432,0.31438,0.31443,0.31448,0.31453,&
                   0.31457,0.31462,0.31466,0.31471,0.31475,0.31479,0.31484,0.31488,&
                   0.31493,0.31498,0.31503,0.31508,0.31514,0.31520,0.31526,0.31532,&
                   0.31538,0.31545,0.31553,0.31561,0.31569,0.31577,0.31587,0.31597,&
                   0.31607,0.31618,0.31630,0.31643,0.31656,0.31670,0.31684,0.31698,&
                   0.31713,0.31729,0.31744,0.31760,0.31777,0.31793,0.31810,0.31827,&
                   0.31844,0.31862,0.31879,0.31897,0.31915,0.31932,0.31950,0.31968,&
                   0.31986,0.32004,0.32022,0.32039,0.32056,0.32073,0.32089,0.32104,&
                   0.32118,0.32131,0.32142,0.32152,0.32159,0.32163,0.32164,0.32160,&
                   0.32152,0.32137,0.32114,0.32081,0.32036,0.31976,0.31900,0.31804,&
                   0.31686,0.31544,0.31378,0.31186,0.30970,0.30731,0.30474,0.30200,&
                   0.29913,0.29617,0.29312,0.29001,0.28686,0.28367,0.28048,0.27738 /)
  T(   1: 160) = (/0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,&
                   0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,&
                   0.00000,0.00001,0.00008,0.00010,0.00010,0.00011,0.00012,0.00013,&
                   0.00015,0.00018,0.00022,0.00027,0.00033,0.00040,0.00047,0.00054,&
                   0.00058,0.00061,0.00063,0.00064,0.00064,0.00064,0.00065,0.00066,&
                   0.00070,0.00075,0.00081,0.00089,0.00097,0.00106,0.00114,0.00123,&
                   0.00131,0.00139,0.00147,0.00154,0.00161,0.00169,0.00176,0.00183,&
                   0.00190,0.00196,0.00201,0.00206,0.00211,0.00217,0.00225,0.00235,&
                   0.00246,0.00258,0.00270,0.00280,0.00290,0.00298,0.00307,0.00317,&
                   0.00331,0.00348,0.00368,0.00393,0.00420,0.00448,0.00476,0.00504,&
                   0.00530,0.00558,0.00586,0.00616,0.00650,0.00689,0.00732,0.00780,&
                   0.00832,0.00889,0.00950,0.01013,0.01079,0.01145,0.01212,0.01278,&
                   0.01342,0.01406,0.01468,0.01531,0.01596,0.01664,0.01733,0.01802,&
                   0.01871,0.01938,0.02007,0.02078,0.02153,0.02229,0.02300,0.02358,&
                   0.02393,0.02400,0.02386,0.02369,0.02363,0.02365,0.02353,0.02297,&
                   0.02175,0.01984,0.01732,0.01439,0.01141,0.00879,0.00685,0.00561,&
                   0.00493,0.00462,0.00451,0.00450,0.00453,0.00460,0.00467,0.00475,&
                   0.00482,0.00489,0.00496,0.00501,0.00505,0.00509,0.00511,0.00514,&
                   0.00516,0.00518,0.00521,0.00524,0.00528,0.00532,0.00537,0.00543,&
                   0.00550,0.00557,0.00565,0.00574,0.00582,0.00591,0.00599,0.00606 /)
  T( 161: 320) = (/0.00613,0.00619,0.00624,0.00628,0.00632,0.00636,0.00640,0.00644,&
                   0.00647,0.00651,0.00655,0.00658,0.00661,0.00665,0.00668,0.00672,&
                   0.00676,0.00680,0.00685,0.00691,0.00698,0.00706,0.00714,0.00723,&
                   0.00732,0.00741,0.00750,0.00758,0.00765,0.00772,0.00777,0.00782,&
                   0.00785,0.00787,0.00787,0.00787,0.00789,0.00796,0.00808,0.00828,&
                   0.00853,0.00879,0.00901,0.00915,0.00920,0.00920,0.00919,0.00920,&
                   0.00923,0.00929,0.00938,0.00947,0.00958,0.00969,0.00980,0.00991,&
                   0.01003,0.01015,0.01027,0.01039,0.01051,0.01063,0.01076,0.01089,&
                   0.01102,0.01116,0.01130,0.01144,0.01159,0.01175,0.01190,0.01207,&
                   0.01223,0.01240,0.01257,0.01275,0.01292,0.01311,0.01329,0.01347,&
                   0.01366,0.01384,0.01403,0.01421,0.01439,0.01457,0.01474,0.01491,&
                   0.01507,0.01522,0.01537,0.01551,0.01565,0.01577,0.01590,0.01602,&
                   0.01613,0.01625,0.01636,0.01648,0.01660,0.01673,0.01686,0.01701,&
                   0.01716,0.01733,0.01750,0.01769,0.01789,0.01810,0.01832,0.01856,&
                   0.01880,0.01904,0.01929,0.01955,0.01980,0.02005,0.02029,0.02052,&
                   0.02074,0.02096,0.02116,0.02134,0.02152,0.02168,0.02183,0.02196,&
                   0.02209,0.02220,0.02231,0.02242,0.02251,0.02261,0.02270,0.02279,&
                   0.02288,0.02298,0.02307,0.02316,0.02326,0.02336,0.02346,0.02356,&
                   0.02366,0.02376,0.02387,0.02397,0.02407,0.02418,0.02428,0.02439,&
                   0.02449,0.02460,0.02471,0.02481,0.02492,0.02503,0.02513,0.02524 /)
  T( 321: 480) = (/0.02535,0.02546,0.02557,0.02568,0.02579,0.02590,0.02601,0.02612,&
                   0.02624,0.02635,0.02647,0.02658,0.02670,0.02682,0.02695,0.02707,&
                   0.02720,0.02733,0.02746,0.02760,0.02773,0.02787,0.02801,0.02815,&
                   0.02829,0.02843,0.02858,0.02872,0.02886,0.02900,0.02914,0.02928,&
                   0.02942,0.02956,0.02970,0.02983,0.02996,0.03009,0.03021,0.03034,&
                   0.03045,0.03057,0.03068,0.03078,0.03088,0.03098,0.03107,0.03115,&
                   0.03123,0.03131,0.03138,0.03144,0.03151,0.03156,0.03162,0.03167,&
                   0.03172,0.03177,0.03182,0.03187,0.03192,0.03197,0.03202,0.03207,&
                   0.03212,0.03218,0.03222,0.03226,0.03228,0.03229,0.03227,0.03223,&
                   0.03218,0.03212,0.03207,0.03206,0.03210,0.03222,0.03242,0.03269,&
                   0.03300,0.03330,0.03355,0.03369,0.03373,0.03367,0.03355,0.03341,&
                   0.03328,0.03320,0.03315,0.03314,0.03316,0.03321,0.03326,0.03332,&
                   0.03339,0.03346,0.03353,0.03360,0.03366,0.03373,0.03380,0.03386,&
                   0.03393,0.03399,0.03405,0.03411,0.03417,0.03423,0.03429,0.03434,&
                   0.03439,0.03444,0.03449,0.03453,0.03457,0.03462,0.03465,0.03469,&
                   0.03473,0.03476,0.03480,0.03483,0.03486,0.03489,0.03492,0.03494,&
                   0.03497,0.03499,0.03501,0.03503,0.03505,0.03507,0.03508,0.03510,&
                   0.03511,0.03513,0.03514,0.03515,0.03517,0.03518,0.03520,0.03521,&
                   0.03523,0.03525,0.03527,0.03529,0.03531,0.03533,0.03535,0.03537,&
                   0.03539,0.03542,0.03544,0.03546,0.03549,0.03551,0.03553,0.03555 /)
  T( 481: 640) = (/0.03557,0.03558,0.03560,0.03562,0.03564,0.03566,0.03567,0.03569,&
                   0.03571,0.03573,0.03575,0.03578,0.03580,0.03583,0.03586,0.03590,&
                   0.03593,0.03597,0.03601,0.03604,0.03608,0.03611,0.03614,0.03617,&
                   0.03621,0.03624,0.03628,0.03632,0.03636,0.03641,0.03646,0.03652,&
                   0.03658,0.03664,0.03670,0.03676,0.03683,0.03689,0.03696,0.03703,&
                   0.03709,0.03716,0.03722,0.03729,0.03735,0.03741,0.03746,0.03752,&
                   0.03757,0.03762,0.03766,0.03770,0.03774,0.03778,0.03782,0.03785,&
                   0.03788,0.03791,0.03793,0.03794,0.03795,0.03795,0.03795,0.03795,&
                   0.03794,0.03792,0.03791,0.03790,0.03789,0.03787,0.03787,0.03786,&
                   0.03785,0.03785,0.03785,0.03786,0.03786,0.03787,0.03788,0.03789,&
                   0.03791,0.03792,0.03794,0.03795,0.03797,0.03799,0.03801,0.03803,&
                   0.03805,0.03807,0.03809,0.03811,0.03813,0.03815,0.03817,0.03820,&
                   0.03822,0.03824,0.03826,0.03828,0.03830,0.03832,0.03834,0.03836,&
                   0.03838,0.03840,0.03842,0.03843,0.03845,0.03847,0.03848,0.03849,&
                   0.03851,0.03852,0.03853,0.03854,0.03854,0.03855,0.03855,0.03855,&
                   0.03855,0.03854,0.03854,0.03852,0.03851,0.03849,0.03848,0.03846,&
                   0.03843,0.03841,0.03839,0.03836,0.03833,0.03830,0.03828,0.03825,&
                   0.03822,0.03819,0.03817,0.03814,0.03812,0.03809,0.03807,0.03805,&
                   0.03804,0.03803,0.03802,0.03801,0.03801,0.03801,0.03802,0.03803,&
                   0.03805,0.03807,0.03810,0.03813,0.03817,0.03821,0.03826,0.03831 /)
  T( 641: 800) = (/0.03836,0.03842,0.03849,0.03855,0.03862,0.03869,0.03876,0.03884,&
                   0.03891,0.03898,0.03906,0.03913,0.03920,0.03926,0.03933,0.03939,&
                   0.03945,0.03951,0.03956,0.03961,0.03966,0.03970,0.03974,0.03976,&
                   0.03978,0.03979,0.03979,0.03977,0.03975,0.03971,0.03967,0.03961,&
                   0.03956,0.03950,0.03943,0.03937,0.03930,0.03924,0.03917,0.03911,&
                   0.03904,0.03897,0.03890,0.03884,0.03877,0.03870,0.03864,0.03858,&
                   0.03852,0.03846,0.03841,0.03837,0.03833,0.03830,0.03827,0.03825,&
                   0.03824,0.03823,0.03823,0.03824,0.03826,0.03829,0.03833,0.03837,&
                   0.03842,0.03848,0.03855,0.03862,0.03870,0.03878,0.03887,0.03897,&
                   0.03906,0.03916,0.03927,0.03937,0.03948,0.03959,0.03969,0.03980,&
                   0.03991,0.04001,0.04011,0.04021,0.04031,0.04041,0.04051,0.04060,&
                   0.04069,0.04079,0.04088,0.04097,0.04106,0.04114,0.04122,0.04129,&
                   0.04136,0.04142,0.04148,0.04152,0.04156,0.04160,0.04163,0.04165,&
                   0.04167,0.04169,0.04170,0.04171,0.04171,0.04172,0.04172,0.04172,&
                   0.04172,0.04172,0.04172,0.04171,0.04171,0.04170,0.04169,0.04169,&
                   0.04168,0.04167,0.04166,0.04165,0.04164,0.04163,0.04162,0.04161,&
                   0.04160,0.04159,0.04159,0.04160,0.04161,0.04163,0.04166,0.04170,&
                   0.04176,0.04184,0.04193,0.04205,0.04219,0.04234,0.04251,0.04268,&
                   0.04283,0.04297,0.04307,0.04313,0.04314,0.04310,0.04302,0.04292,&
                   0.04285,0.04283,0.04291,0.04315,0.04360,0.04432,0.04537,0.04677 /)
  Q(   1: 160) = (/0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,&
                   0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,&
                   0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00010,&
                   0.00034,0.00016,0.00012,0.00011,0.00011,0.00011,0.00011,0.00011,&
                   0.00011,0.00012,0.00013,0.00014,0.00016,0.00018,0.00019,0.00020,&
                   0.00022,0.00024,0.00027,0.00030,0.00036,0.00042,0.00050,0.00058,&
                   0.00066,0.00074,0.00081,0.00087,0.00093,0.00099,0.00104,0.00110,&
                   0.00115,0.00120,0.00124,0.00128,0.00131,0.00135,0.00139,0.00144,&
                   0.00150,0.00157,0.00166,0.00175,0.00186,0.00197,0.00209,0.00222,&
                   0.00237,0.00255,0.00274,0.00297,0.00321,0.00348,0.00375,0.00402,&
                   0.00431,0.00461,0.00495,0.00534,0.00580,0.00635,0.00698,0.00770,&
                   0.00852,0.00941,0.01038,0.01141,0.01248,0.01362,0.01480,0.01603,&
                   0.01730,0.01860,0.01992,0.02124,0.02258,0.02394,0.02536,0.02685,&
                   0.02840,0.03004,0.03171,0.03339,0.03501,0.03650,0.03780,0.03885,&
                   0.03955,0.03988,0.03993,0.03994,0.04005,0.04015,0.03983,0.03860,&
                   0.03619,0.03261,0.02807,0.02288,0.01758,0.01285,0.00919,0.00671,&
                   0.00523,0.00444,0.00406,0.00391,0.00386,0.00387,0.00390,0.00394,&
                   0.00400,0.00406,0.00412,0.00418,0.00423,0.00428,0.00433,0.00437,&
                   0.00440,0.00444,0.00447,0.00450,0.00453,0.00456,0.00459,0.00463,&
                   0.00467,0.00473,0.00480,0.00489,0.00498,0.00507,0.00515,0.00522 /)
  Q( 161: 320) = (/0.00527,0.00530,0.00531,0.00532,0.00533,0.00535,0.00538,0.00543,&
                   0.00549,0.00556,0.00565,0.00573,0.00582,0.00591,0.00599,0.00608,&
                   0.00616,0.00625,0.00633,0.00642,0.00652,0.00662,0.00672,0.00682,&
                   0.00693,0.00704,0.00716,0.00728,0.00741,0.00753,0.00766,0.00776,&
                   0.00785,0.00792,0.00797,0.00801,0.00806,0.00817,0.00835,0.00862,&
                   0.00897,0.00935,0.00971,0.00999,0.01019,0.01032,0.01044,0.01058,&
                   0.01074,0.01093,0.01115,0.01138,0.01162,0.01187,0.01212,0.01238,&
                   0.01264,0.01290,0.01316,0.01343,0.01370,0.01397,0.01425,0.01452,&
                   0.01480,0.01508,0.01536,0.01564,0.01592,0.01620,0.01648,0.01677,&
                   0.01707,0.01737,0.01767,0.01798,0.01830,0.01863,0.01895,0.01929,&
                   0.01963,0.01997,0.02032,0.02067,0.02103,0.02140,0.02177,0.02216,&
                   0.02255,0.02295,0.02336,0.02377,0.02420,0.02463,0.02507,0.02551,&
                   0.02596,0.02641,0.02686,0.02731,0.02775,0.02819,0.02863,0.02906,&
                   0.02948,0.02989,0.03029,0.03067,0.03104,0.03140,0.03175,0.03210,&
                   0.03243,0.03277,0.03311,0.03344,0.03378,0.03412,0.03447,0.03483,&
                   0.03519,0.03556,0.03595,0.03634,0.03675,0.03716,0.03758,0.03801,&
                   0.03845,0.03888,0.03932,0.03976,0.04019,0.04062,0.04104,0.04145,&
                   0.04186,0.04225,0.04264,0.04301,0.04337,0.04373,0.04407,0.04441,&
                   0.04475,0.04508,0.04541,0.04573,0.04605,0.04637,0.04669,0.04701,&
                   0.04732,0.04763,0.04794,0.04824,0.04855,0.04885,0.04914,0.04943 /)
  Q( 321: 480) = (/0.04972,0.05001,0.05029,0.05057,0.05084,0.05110,0.05136,0.05162,&
                   0.05187,0.05211,0.05236,0.05259,0.05283,0.05306,0.05328,0.05350,&
                   0.05372,0.05393,0.05414,0.05434,0.05453,0.05472,0.05490,0.05508,&
                   0.05526,0.05543,0.05560,0.05576,0.05593,0.05609,0.05625,0.05641,&
                   0.05657,0.05673,0.05688,0.05704,0.05720,0.05736,0.05752,0.05769,&
                   0.05785,0.05802,0.05819,0.05837,0.05854,0.05872,0.05890,0.05909,&
                   0.05928,0.05947,0.05966,0.05986,0.06006,0.06027,0.06047,0.06068,&
                   0.06089,0.06110,0.06131,0.06152,0.06173,0.06193,0.06213,0.06231,&
                   0.06248,0.06263,0.06275,0.06284,0.06290,0.06291,0.06288,0.06282,&
                   0.06273,0.06264,0.06259,0.06262,0.06277,0.06308,0.06356,0.06417,&
                   0.06484,0.06549,0.06602,0.06637,0.06650,0.06644,0.06625,0.06600,&
                   0.06576,0.06556,0.06542,0.06533,0.06528,0.06526,0.06525,0.06525,&
                   0.06525,0.06525,0.06525,0.06526,0.06526,0.06527,0.06527,0.06528,&
                   0.06529,0.06531,0.06533,0.06536,0.06539,0.06542,0.06546,0.06551,&
                   0.06556,0.06562,0.06568,0.06575,0.06582,0.06590,0.06599,0.06608,&
                   0.06617,0.06626,0.06636,0.06647,0.06657,0.06668,0.06679,0.06690,&
                   0.06702,0.06713,0.06725,0.06737,0.06749,0.06760,0.06772,0.06784,&
                   0.06796,0.06808,0.06819,0.06831,0.06842,0.06853,0.06863,0.06873,&
                   0.06884,0.06894,0.06903,0.06913,0.06923,0.06933,0.06942,0.06951,&
                   0.06960,0.06968,0.06976,0.06983,0.06990,0.06996,0.07002,0.07007 /)
  Q( 481: 640) = (/0.07011,0.07015,0.07019,0.07023,0.07026,0.07028,0.07031,0.07033,&
                   0.07035,0.07037,0.07039,0.07041,0.07043,0.07046,0.07049,0.07052,&
                   0.07056,0.07060,0.07064,0.07067,0.07071,0.07074,0.07078,0.07081,&
                   0.07084,0.07087,0.07090,0.07093,0.07097,0.07100,0.07104,0.07109,&
                   0.07113,0.07117,0.07122,0.07126,0.07131,0.07135,0.07139,0.07143,&
                   0.07147,0.07151,0.07154,0.07157,0.07160,0.07162,0.07163,0.07164,&
                   0.07164,0.07163,0.07162,0.07159,0.07154,0.07149,0.07142,0.07133,&
                   0.07123,0.07112,0.07100,0.07088,0.07074,0.07060,0.07046,0.07032,&
                   0.07018,0.07005,0.06993,0.06982,0.06972,0.06963,0.06956,0.06950,&
                   0.06946,0.06944,0.06943,0.06943,0.06945,0.06948,0.06952,0.06956,&
                   0.06962,0.06969,0.06976,0.06984,0.06992,0.07001,0.07010,0.07020,&
                   0.07030,0.07040,0.07051,0.07062,0.07073,0.07084,0.07095,0.07106,&
                   0.07118,0.07129,0.07140,0.07152,0.07163,0.07174,0.07184,0.07195,&
                   0.07205,0.07215,0.07225,0.07234,0.07243,0.07251,0.07259,0.07266,&
                   0.07273,0.07279,0.07284,0.07289,0.07293,0.07296,0.07297,0.07298,&
                   0.07297,0.07296,0.07293,0.07289,0.07283,0.07277,0.07270,0.07262,&
                   0.07253,0.07243,0.07232,0.07221,0.07210,0.07198,0.07185,0.07172,&
                   0.07159,0.07146,0.07132,0.07119,0.07106,0.07092,0.07079,0.07067,&
                   0.07055,0.07043,0.07032,0.07021,0.07011,0.07001,0.06993,0.06985,&
                   0.06978,0.06971,0.06965,0.06960,0.06956,0.06952,0.06949,0.06946 /)
  Q( 641: 800) = (/0.06944,0.06942,0.06941,0.06940,0.06939,0.06939,0.06938,0.06939,&
                   0.06939,0.06939,0.06940,0.06941,0.06942,0.06943,0.06944,0.06946,&
                   0.06947,0.06949,0.06951,0.06953,0.06955,0.06958,0.06961,0.06964,&
                   0.06967,0.06971,0.06974,0.06978,0.06983,0.06987,0.06992,0.06997,&
                   0.07002,0.07007,0.07013,0.07018,0.07024,0.07029,0.07035,0.07041,&
                   0.07047,0.07053,0.07059,0.07065,0.07071,0.07077,0.07083,0.07089,&
                   0.07095,0.07100,0.07106,0.07112,0.07117,0.07122,0.07128,0.07133,&
                   0.07138,0.07142,0.07147,0.07151,0.07155,0.07159,0.07163,0.07166,&
                   0.07169,0.07172,0.07175,0.07178,0.07181,0.07184,0.07186,0.07189,&
                   0.07191,0.07193,0.07195,0.07197,0.07199,0.07201,0.07202,0.07203,&
                   0.07205,0.07206,0.07207,0.07207,0.07208,0.07208,0.07209,0.07209,&
                   0.07209,0.07208,0.07208,0.07208,0.07207,0.07206,0.07205,0.07204,&
                   0.07203,0.07202,0.07201,0.07200,0.07199,0.07198,0.07197,0.07195,&
                   0.07194,0.07193,0.07191,0.07190,0.07188,0.07187,0.07185,0.07183,&
                   0.07182,0.07180,0.07178,0.07176,0.07175,0.07173,0.07171,0.07169,&
                   0.07167,0.07165,0.07163,0.07161,0.07160,0.07158,0.07156,0.07155,&
                   0.07153,0.07152,0.07151,0.07151,0.07152,0.07153,0.07156,0.07160,&
                   0.07166,0.07174,0.07185,0.07199,0.07216,0.07237,0.07260,0.07286,&
                   0.07312,0.07339,0.07363,0.07383,0.07398,0.07407,0.07410,0.07410,&
                   0.07409,0.07411,0.07422,0.07446,0.07489,0.07557,0.07654,0.07780 /)

  endif
! irun = irun+1

  energy(1:nbin) = e(1:nbin)
  select case (pattyp)
    case (1)
      patfraction(1:nbin) = s(1:nbin)
    case (2)
      patfraction(1:nbin) = d(1:nbin)
    case (3)
      patfraction(1:nbin) = t(1:nbin)
    case (4)
      patfraction(1:nbin) = q(1:nbin)
    case default
      patfraction(1:nbin) = 0
  end select
  if (maxbin > nbin) then
    energy(nbin+1:maxbin) = 0.0
    patfraction(nbin+1:maxbin) = patfraction(nbin)
  endif

end subroutine patmod_mos


end module epatplot_mod
